# -*- coding: utf-8 -*-
"""
This module provides code for decoding and deserialising binary ROS messages
into Python data structures.
"""
from typing import (Optional, Iterator, Callable, Any, List, Type, TypeVar,
                    BinaryIO)
import functools
import struct

from .base import Time, Duration

T = TypeVar('T')

_SIMPLE_TO_STRUCT = {
    'int8': 'b',
    'uint8': 'B',
    'bool': 'B',
    'int16': 'h',
    'uint16': 'H',
    'int32': 'i',
    'uint32': 'I',
    'int64': 'q',
    'uint64': 'Q',
    'float32': 'f',
    'float64': 'd',
    # deprecated types
    'char': 'B',  # unsigned
    'byte': 'b'  # signed
}


def is_simple(typ: str) -> bool:
    """Determines whether a given type is a simple type."""
    return typ in _SIMPLE_TO_STRUCT


def get_pattern(typ: str) -> str:
    """Returns the struct pattern for a given simple type."""
    return _SIMPLE_TO_STRUCT[typ]


def simple_decoder(typ: str) -> Callable[[bytes], Any]:
    """Returns a decoder for a specified simple type."""
    pattern = '<' + get_pattern(typ)

    def decoder(v: bytes) -> Any:
        return struct.unpack(pattern, v)[0]

    def bool_decoder(v: bytes) -> bool:
        return bool(struct.unpack(pattern, v)[0])

    return bool_decoder if typ == 'bool' else decoder


def simple_reader(typ: str) -> Callable[[BinaryIO], Any]:
    """Returns a reader for a specified simple type."""
    pattern = '<' + get_pattern(typ)
    decoder = simple_decoder(typ)
    size = struct.calcsize(pattern)

    def reader(b: BinaryIO) -> Any:
        return decoder(b.read(size))

    return reader


decode_int8 = simple_decoder('int8')
decode_uint8 = simple_decoder('uint8')
decode_int16 = simple_decoder('int16')
decode_uint16 = simple_decoder('uint16')
decode_int32 = simple_decoder('int32')
decode_uint32 = simple_decoder('uint32')
decode_int64 = simple_decoder('int64')
decode_uint64 = simple_decoder('uint64')
decode_float32 = simple_decoder('float32')
decode_float64 = simple_decoder('float64')
decode_char = simple_decoder('char')
decode_byte = simple_decoder('byte')
decode_bool = simple_decoder('bool')

read_int8 = simple_reader('int8')
read_uint8 = simple_reader('uint8')
read_int16 = simple_reader('int16')
read_uint16 = simple_reader('uint16')
read_int32 = simple_reader('int32')
read_uint32 = simple_reader('uint32')
read_int64 = simple_reader('int64')
read_uint64 = simple_reader('uint64')
read_float32 = simple_reader('float32')
read_float64 = simple_reader('float64')
read_char = simple_reader('char')
read_byte = simple_reader('byte')
read_bool = simple_reader('bool')


def decode_time(b: bytes) -> Time:
    secs = decode_uint32(b[0:4])
    nsecs = decode_uint32(b[4:8])
    return Time(secs, nsecs)


def read_time(b: BinaryIO) -> Time:
    return decode_time(b.read(8))


def decode_duration(b: bytes) -> Duration:
    secs = decode_uint32(b[0:4])
    nsecs = decode_uint32(b[4:8])
    return Duration(secs, nsecs)


def read_duration(b: BinaryIO) -> Duration:
    return decode_duration(b.read(8))


def decode_string(b: bytes) -> str:
    return b.decode('utf-8')


def read_fixed_length_string(size: int, b: BinaryIO) -> str:
    """Reads a fixed-length string from a bytestream."""
    return decode_string(b.read(size))


def read_string(b: BinaryIO) -> str:
    """Reads a variable-length string from a bytestream."""
    size = read_uint32(b)
    return read_fixed_length_string(size, b)


def string_reader(length: Optional[int] = None
                  ) -> Callable[[BinaryIO], str]:
    """Returns a reader for (possibly fixed-length) strings."""
    if length is None:
        return read_string
    else:
        return functools.partial(read_fixed_length_string, length)


def simple_array_reader(typ: str,
                        length: Optional[int] = None
                        ) -> Callable[[BinaryIO], List[Any]]:
    """Returns a reader for a simple array."""
    base_pattern = get_pattern(typ)

    # fixed length: precompute pattern
    if length is not None:
        pattern = f'<{length}{base_pattern}'
        size = struct.calcsize(pattern)

        def fixed_reader(b: BinaryIO) -> List[Any]:
            return list(struct.unpack(pattern, b.read(size)))

        return fixed_reader

    # variable length
    def var_reader(b: BinaryIO) -> List[Any]:
        length = read_uint32(b)
        pattern = f'<{length}{base_pattern}'
        size = struct.calcsize(pattern)
        return list(struct.unpack(pattern, b.read(size)))

    return var_reader


def complex_array_reader(factory: Callable[[BinaryIO], T],
                         length: Optional[int] = None
                         ) -> Callable[[BinaryIO], List[T]]:
    """Returns a reader for a complex array."""
    def read_fixed(length: int, b: BinaryIO) -> List[T]:
        return [factory(b) for i in range(length)]

    def read_var(b: BinaryIO) -> List[T]:
        length = read_uint32(b)
        return read_fixed(length, b)

    if length is None:
        return read_var
    else:
        return functools.partial(read_fixed, length)



# -*- coding: utf-8 -*-
"""
This module provides a bag file writer that writes the contents of a bag to a
binary file on disk.

See
---
https://github.com/ros/ros_comm/blob/melodic-devel/tools/rosbag/src/rosbag/bag.py
"""
__all__ = ('BagWriter',)

from typing import BinaryIO, Iterable, Dict, Type, Tuple
import logging

from .core import (BagMessage, OpCode, Compression, ConnectionInfo, Chunk,
                   Index, IndexEntry, ChunkConnection)
from ..definitions import Message
from ..definitions.encode import *

BIN_CHUNK_INFO_VERSION = encode_uint32(1)
BIN_INDEX_VERSION = encode_uint32(1)

logger: logging.Logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class BagWriter:
    """
    Provides an interface for writing messages to a bag file on disk.

    Attributes
    ----------
    filename: str
        The name of the file to which the bag will be written.
    """
    def __init__(self, fn: str) -> None:
        self.__fn = fn
        self.__fp: BinaryIO = open(fn, 'wb')
        self.__connections: Dict[str, ConnectionInfo] = {}
        self.__chunks: List[Chunk] = []
        self.__pos_header = 0
        self.__pos_chunks = 0
        self.__pos_index = 0

    @property
    def filename(self) -> str:
        return self.__fn

    def _write_header(self,
                      code: Optional[OpCode],
                      fields: Dict[str, bytes]
                      ) -> None:
        # write a length of zero for now and correct that length once the
        # fields have been written
        pos_size = self.__fp.tell()
        write_uint32(0, self.__fp)
        pos_content = self.__fp.tell()

        if code:
            fields['op'] = code.value
        for name, bin_value in fields.items():
            bin_name = f'{name}='.encode('utf-8')
            size_field = len(bin_name) + len(bin_value)
            write_uint32(size_field, self.__fp)
            self.__fp.write(bin_name)
            self.__fp.write(bin_value)

        # correct the length
        pos_end = self.__fp.tell()
        size_total = pos_end - pos_content
        self.__fp.seek(pos_size)
        write_uint32(size_total, self.__fp)
        self.__fp.seek(pos_end)

    def _write_header_record(self) -> None:
        logger.debug("writing bag header record")
        self.__fp.seek(self.__pos_header)
        self._write_header(OpCode.HEADER, {
            'chunk_count': encode_uint32(len(self.__chunks)),
            'conn_count': encode_uint32(len(self.__connections)),
            'index_pos': encode_uint64(self.__pos_index)
        })

        # ensure the bag header record is 4096 characters long by padding it
        # with ASCII space characters (0x20) where necessary.
        #
        # the ROS bag format doesn't consider size fields when computing the
        # length of the header and the record
        pos_current = self.__fp.tell()
        size = 4096
        size_header = pos_current - self.__pos_header - 4
        size_padding = size - size_header

        write_uint32(size_padding, self.__fp)
        padding = b'\x20' * size_padding
        self.__fp.write(padding)
        logger.debug("wrote bag header record")

    def _write_message(self,
                       pos_chunk_start: int,
                       index: Index,
                       message: BagMessage
                       ) -> None:
        typ = message.message.__class__
        connection = self._get_connection(message.topic, typ)
        logger.debug("writing message on connection [%s]", connection.topic)

        pos_header = self.__fp.tell()
        self._write_header(OpCode.MESSAGE_DATA, {
            'conn': encode_uint32(connection.conn),
            'time': encode_time(message.time)})

        bin_data = message.message.encode()
        size_data = len(bin_data)
        write_uint32(size_data, self.__fp)
        self.__fp.write(bin_data)

        # update index
        offset = pos_header - pos_chunk_start
        index_entry = IndexEntry(time=message.time,
                                 pos=pos_header,
                                 offset=offset)
        if connection.conn not in index:
            index[connection.conn] = []
        index[connection.conn].append(index_entry)

    def _write_chunk_data(self, messages: Iterable[BagMessage]) -> Index:
        logger.debug("writing messages to chunk")
        index: Index = {}
        pos_start = self.__fp.tell()
        for m in messages:
            self._write_message(pos_start, index, m)
        logger.debug("wrote messages to chunk")
        return index

    def _write_chunk_record(self,
                            compression: Compression,
                            messages: Iterable[BagMessage]
                            ) -> Tuple[Chunk, Index]:
        bin_compression = compression.value.encode('utf-8')

        # for now, we write a bogus header and size field
        # once we've finished writing the data, we'll correct them
        pos_header = self.__fp.tell()
        self._write_header(OpCode.CHUNK, {
            'compression': bin_compression,
            'size': encode_uint32(0)})
        write_uint32(0, self.__fp)
        pos_data = self.__fp.tell()

        # write chunk contents
        index = self._write_chunk_data(messages)

        # determine time of earliest and latest message in the bag
        times = [e.time for ci in index.values() for e in ci]
        time_start = min(times)
        time_end = max(times)

        # compute chunk size
        pos_end = self.__fp.tell()
        size_compressed = pos_end - pos_data
        size_uncompressed = size_compressed

        # update header and size
        self.__fp.seek(pos_header)
        self._write_header(OpCode.CHUNK, {
            'compression': bin_compression,
            'size': encode_uint32(size_uncompressed)})
        write_uint32(size_compressed, self.__fp)
        self.__fp.seek(pos_end)

        # build a description of the chunk
        conns = [ChunkConnection(conn, len(entries))
                 for conn, entries in index.items()]
        chunk = Chunk(pos_record=pos_header,  # type: ignore
                      pos_data=pos_data,
                      time_start=time_start,
                      time_end=time_end,
                      connections=conns,
                      compression=compression,
                      size_compressed=size_compressed,
                      size_uncompressed=size_uncompressed)
        return chunk, index

    def _write_connection_index(self,
                                conn: int,
                                entries: List[IndexEntry]
                                ) -> None:
        logger.debug("writing connection index [%d]", conn)
        num_entries = len(entries)
        size_data = num_entries * 12
        self._write_header(OpCode.INDEX_DATA, {
            'ver': BIN_INDEX_VERSION,
            'conn': encode_uint32(conn),
            'count': encode_uint32(num_entries)})
        write_uint32(size_data, self.__fp)
        for entry in entries:
            write_time(entry.time, self.__fp)
            write_uint32(entry.offset, self.__fp)

    def _write_chunk(self,
                     compression: Compression,
                     messages: Iterable[BagMessage]
                     ) -> None:
        # TODO for now, we only support uncompressed writing
        assert compression == Compression.NONE
        chunk, index = self._write_chunk_record(compression, messages)
        self.__chunks.append(chunk)
        for conn, entries in index.items():
            self._write_connection_index(conn, entries)

    def _write_connection_record(self, conn: ConnectionInfo) -> None:
        self._write_header(OpCode.CONNECTION_INFO, {
            'conn': encode_uint32(conn.conn),
            'topic': conn.topic.encode('utf-8')})

        # write the connection header
        header_conn: Dict[str, bytes] = {}
        if conn.callerid is not None:
            header_conn['callerid'] = conn.callerid.encode('utf-8')
        if conn.latching is not None:
            header_conn['latching'] = conn.latching.encode('utf-8')
        header_conn['topic'] = conn.topic_original.encode('utf-8')
        header_conn['type'] = conn.typ.encode('utf-8')
        header_conn['md5sum'] = conn.md5sum.encode('utf-8')
        header_conn['message_definition'] = \
            conn.message_definition.encode('utf-8')
        self._write_header(None, header_conn)

    def _get_connection(self,
                        topic: str,
                        typ: Type[Message]
                        ) -> ConnectionInfo:
        # if there isn't a connection for the topic, create one and write a
        # connection record.
        if topic not in self.__connections:
            msg_format = typ.format
            conn = len(self.__connections)
            info = ConnectionInfo(conn=conn,
                                  topic=topic,
                                  topic_original=topic,
                                  typ=msg_format.fullname,
                                  md5sum=typ.md5sum(),
                                  message_definition=msg_format.definition,
                                  callerid=None,
                                  latching=None)
            self.__connections[topic] = info
            self._write_connection_record(info)

        return self.__connections[topic]

    def _write_chunk_info_record(self, chunk: Chunk) -> None:
        num_connections = len(chunk.connections)
        pos_header = self.__fp.tell()
        self._write_header(OpCode.CHUNK_INFO, {
            'ver': BIN_CHUNK_INFO_VERSION,
            'chunk_pos': encode_uint64(chunk.pos_record),
            'start_time': encode_time(chunk.time_start),
            'end_time': encode_time(chunk.time_end),
            'count': encode_uint32(num_connections)})

        size_data = num_connections * 8
        write_uint32(size_data, self.__fp)
        for connection in chunk.connections:
            write_uint32(connection.uid, self.__fp)
            write_uint32(connection.count, self.__fp)

    def _write_index(self) -> None:
        for connection in self.__connections.values():
            self._write_connection_record(connection)
        for chunk in self.__chunks:
            self._write_chunk_info_record(chunk)

    def write(self, messages: Iterable[BagMessage]) -> None:
        """
        Writes a sequence of messages to the bag.
        Any existing bag file contents will be overwritten.
        """
        self.__fp.truncate(0)
        self.__fp.write('#ROSBAG V2.0\n'.encode('utf-8'))

        # create a placeholder header for now
        self.__pos_header = self.__fp.tell()
        self._write_header_record()

        # for now, we write to a single, uncompressed chunk
        # each chunk record is followed by a sequence of IndexData record
        # - each connection in the chunk is represented by an IndexData record
        self.__pos_chunks = self.__fp.tell()
        self._write_chunk(Compression.NONE, messages)

        # write index
        self.__pos_index = self.__fp.tell()
        self._write_index()

        # fix the header
        self._write_header_record()

    def close(self) -> None:
        self.__fp.close()



# -*- coding: utf-8 -*-
__all__ = (
    'ShellProxy',
    'ServiceManagerProxy',
    'ParameterServerProxy',
    'NodeManagerProxy',
    'NodeProxy',
    'ROSProxy',
    'BagRecorderProxy',
    'BagPlayerProxy',
    'FileProxy',
    'ContainerProxy',
    'CatkinProxy',
    'CatkinToolsProxy',
    'CatkinMakeProxy',
    'ContainerProxyManager'
)

from typing import (Tuple, Dict, Optional, Iterator, Any, List, Union,
                    Collection)
import os
import xmlrpc.client
import shlex
import logging
import time

from .shell import ShellProxy
from .file import FileProxy
from .catkin import CatkinProxy, CatkinToolsProxy, CatkinMakeProxy
from .container import ContainerProxy, ContainerProxyManager
from .parameters import ParameterServerProxy
from .bag import BagRecorderProxy, BagPlayerProxy
from .node import NodeProxy, NodeManagerProxy
from .service import ServiceManagerProxy
from ..description import SystemDescription
from ..exceptions import ROSWireException

logger: logging.Logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class ROSProxy:
    """Provides access to a remote ROS master via XML-RPC."""
    def __init__(self,
                 description: SystemDescription,
                 shell: ShellProxy,
                 files: FileProxy,
                 ws_host: str,
                 ip_address: str,
                 port: int = 11311
                 ) -> None:
        self.__description = description
        self.__shell = shell
        self.__files = files
        self.__ws_host = ws_host
        self.__caller_id = '/roswire'
        self.__port = port
        self.__ip_address = ip_address
        self.__uri = f"http://{ip_address}:{port}"
        logger.debug("connecting to ROS Master: %s", self.__uri)
        self.__connection = xmlrpc.client.ServerProxy(self.__uri)
        time.sleep(5)  # FIXME #1
        self.__parameters = ParameterServerProxy(self.__connection)
        self.__nodes: NodeManagerProxy = \
            NodeManagerProxy(self.__ip_address,
                             self.__connection,
                             self.__shell)
        self.__services: ServiceManagerProxy = \
            ServiceManagerProxy(self.__description,
                                self.__ip_address,
                                self.__connection,
                                self.__shell)

    @property
    def uri(self) -> str:
        """The URI of the ROS Master."""
        return self.__uri

    @property
    def nodes(self) -> NodeManagerProxy:
        """Provides access to the nodes running on this ROS master."""
        return self.__nodes

    @property
    def services(self) -> ServiceManagerProxy:
        """Provides access to the services advertised on this ROS master."""
        return self.__services

    @property
    def parameters(self) -> ParameterServerProxy:
        """Provides access to the parameter server for this ROS Master."""
        return self.__parameters

    @property
    def connection(self) -> xmlrpc.client.ServerProxy:
        """The XML-RPC connection to the ROS master."""
        return self.__connection

    @property
    def topic_to_type(self) -> Dict[str, str]:
        conn = self.connection
        code, msg, result = conn.getTopicTypes(self.__caller_id)
        if code != 1:
            raise ROSWireException("bad API call!")
        return {name: typ for (name, typ) in result}

    def launch(self,
               filename: str,
               package: Optional[str] = None,
               *,
               args: Optional[Dict[str, Union[int, str]]] = None,
               prefix: Optional[str] = None
               ) -> None:
        """Provides an interface to roslaunch.

        Parameters
        ----------
        filename: str
            The name of the launch file (or an absolute path).
        package: str, optional
            The name of the package to which the launch file belongs.
        args: Dict[str, Union[int, str]], optional
            Keyword arguments that should be supplied to roslaunch.
        prefix: str, optional
            An optional prefix to add before the roslaunch command.
        """
        if not args:
            args = {}
        launch_args: List[str] = [f'{arg}:={val}' for arg, val in args.items()]
        cmd = ['roslaunch']
        if package:
            cmd += [shlex.quote(package)]
        cmd += [shlex.quote(filename)]
        cmd += launch_args
        if prefix:
            cmd = [prefix] + cmd
        cmd_str = ' '.join(cmd)
        self.__shell.non_blocking_execute(cmd_str)

    def record(self,
               fn: str,
               exclude_topics: Optional[Collection[str]] = None
               ) -> BagRecorderProxy:
        """Provides an interface to rosbag for recording ROS topics."""
        return BagRecorderProxy(fn,
                                self.__ws_host,
                                self.__shell,
                                self.__nodes,
                                exclude_topics=exclude_topics)

    def playback(self,
                 fn: str,
                 *,
                 file_on_host: bool = True
                 ) -> BagPlayerProxy:
        """Provides an interface to rosbag for replaying bag files."""
        fn_ctr: str
        delete_file_after_use: bool = False
        if file_on_host:
            if fn.startswith(self.__ws_host):
                fn_ctr = os.path.join('/.roswire', fn[len(self.__ws_host):])
            else:
                delete_file_after_use = True
                fn_ctr = self.__files.mktemp(suffix='.bag')
                logger.debug("copying bag from host [%s] to container [%s]",
                             fn, fn_ctr)
                self.__files.copy_from_host(fn, fn_ctr)
        else:
            fn_ctr = fn
        logger.debug("playing back bag file: %s", fn_ctr)
        return BagPlayerProxy(fn_ctr,
                              self.__shell,
                              self.__files,
                              delete_file_after_use=delete_file_after_use)



from typing import Iterator
import contextlib
import tempfile
import shutil
import os

import pytest

from roswire.proxy import FileProxy
from roswire.definitions import MsgFormat, SrvFormat, Package, PackageDatabase

from test_file import build_file_proxy
from test_basic import build_shell_proxy


def test_to_and_from_dict():
    pkg = 'tf'
    msg_tf = MsgFormat.from_dict({
        'package': pkg,
        'name': 'tfMessage',
        'fields': [{'type': 'geometry_msgs/TransformStamped[]',
                    'name': 'transforms'}]})
    srv_fg = SrvFormat.from_dict({
        'package': pkg,
        'name': 'FrameGraph',
        'response': {
            'fields': [{'type': 'string', 'name': 'dot_graph'}]}})
    p = Package(name=pkg,
                path='/ros_ws/src/geometry/tf',
                messages=[msg_tf],
                actions=[],
                services=[srv_fg])
    assert p == Package.from_dict(p.to_dict())


def test_build():
    with build_file_proxy() as files:
        path = '/ros_ws/src/geometry/tf'
        expected = Package.from_dict({
            'path': path,
            'name': 'tf',
            'messages': [
                {'name': 'tfMessage',
                 'fields': [{'type': 'geometry_msgs/TransformStamped[]',
                             'name': 'transforms'}]}
            ],
            'services': [
                {'name': 'FrameGraph',
                 'response': {
                 'fields': [{'type': 'string',
                             'name': 'dot_graph'}]}}
            ]
        })
        actual = Package.build(path, files)
        assert actual == expected


def test_database_paths():
    with build_shell_proxy() as shell:
        expected = [
            '/ros_ws/src/catkin',
            '/ros_ws/src/genmsg',
            '/ros_ws/src/gencpp',
            '/ros_ws/src/genlisp',
            '/ros_ws/src/genpy',
            '/ros_ws/src/cmake_modules',
            '/ros_ws/src/class_loader',
            '/ros_ws/src/roscpp_core/cpp_common',
            '/ros_ws/src/mavlink',
            '/ros_ws/src/mavros/libmavconn',
            '/ros_ws/src/message_generation',
            '/ros_ws/src/message_runtime',
            '/ros_ws/src/orocos_kinematics_dynamics/orocos_kdl',
            '/ros_ws/src/ros/rosbuild',
            '/ros_ws/src/ros/rosclean',
            '/ros_ws/src/roscpp_core/roscpp_traits',
            '/ros_ws/src/ros_comm/rosgraph',
            '/ros_ws/src/ros/roslang',
            '/ros_ws/src/ros_comm/rosmaster',
            '/ros_ws/src/ros_comm/rosmsg',
            '/ros_ws/src/rospack',
            '/ros_ws/src/ros/roslib',
            '/ros_ws/src/ros_comm/rosparam',
            '/ros_ws/src/ros_comm/rospy',
            '/ros_ws/src/ros_comm/rosservice',
            '/ros_ws/src/roscpp_core/rostime',
            '/ros_ws/src/roscpp_core/roscpp_serialization',
            '/ros_ws/src/ros_comm/roslaunch',
            '/ros_ws/src/ros/rosunit',
            '/ros_ws/src/angles',
            '/ros_ws/src/ros_comm/rosconsole',
            '/ros_ws/src/pluginlib',
            '/ros_ws/src/rosconsole_bridge',
            '/ros_ws/src/ros_comm/roslz4',
            '/ros_ws/src/ros_comm/rosbag_storage',
            '/ros_ws/src/ros_comm/rostest',
            '/ros_ws/src/std_msgs',
            '/ros_ws/src/common_msgs/actionlib_msgs',
            '/ros_ws/src/common_msgs/diagnostic_msgs',
            '/ros_ws/src/common_msgs/geometry_msgs',
            '/ros_ws/src/geometry/eigen_conversions',
            '/ros_ws/src/mavros/mavros_msgs',
            '/ros_ws/src/common_msgs/nav_msgs',
            '/ros_ws/src/ros_comm_msgs/rosgraph_msgs',
            '/ros_ws/src/common_msgs/sensor_msgs',
            '/ros_ws/src/ros_comm_msgs/std_srvs',
            '/ros_ws/src/geometry2/tf2_msgs',
            '/ros_ws/src/geometry2/tf2',
            '/ros_ws/src/urdf/urdf_parser_plugin',
            '/ros_ws/src/common_msgs/visualization_msgs',
            '/ros_ws/src/ros_comm/xmlrpcpp',
            '/ros_ws/src/ros_comm/roscpp',
            '/ros_ws/src/ros_comm/rosout',
            '/ros_ws/src/vision_opencv/cv_bridge',
            '/ros_ws/src/diagnostics/diagnostic_updater',
            '/ros_ws/src/ros_comm/message_filters',
            '/ros_ws/src/image_common/image_transport',
            '/ros_ws/src/ros_comm/rosnode',
            '/ros_ws/src/ros_comm/rostopic',
            '/ros_ws/src/ros_comm/roswtf',
            '/ros_ws/src/geometry2/tf2_py',
            '/ros_ws/src/ros_comm/topic_tools',
            '/ros_ws/src/ros_comm/rosbag',
            '/ros_ws/src/actionlib',
            '/ros_ws/src/geometry2/tf2_ros',
            '/ros_ws/src/mavros/mavros',
            '/ros_ws/src/geometry/tf',
            '/ros_ws/src/urdf/urdf',
            '/ros_ws/src/mavros/mavros_extras',
            '/opt/ros/indigo/share',
            '/opt/ros/indigo/stacks'
        ]
        actual = PackageDatabase.paths(shell)
        assert actual == expected


def test_database_from_paths():
    with build_file_proxy() as files:
        paths = [
            '/ros_ws/src/angles',
            '/ros_ws/src/mavros',
            '/ros_ws/src/geometry2/tf2',
            '/ros_ws/src/geometry2/tf2_msgs',
            '/ros_ws/src/geometry2/tf2_py',
            '/ros_ws/src/geometry2/tf2_ros'
        ]
        db = PackageDatabase.from_paths(files, paths)
        assert len(db) == len(paths)
        assert set(db) == {'angles', 'mavros', 'tf2',
                           'tf2_msgs', 'tf2_py', 'tf2_ros'}



# -*- coding: utf-8 -*-
"""
This module provides access to the ROSWire session.
"""
__all__ = ('ROSWire',)

from typing import Optional, Dict, Iterator
from uuid import uuid4
import os
import pathlib
import logging
import contextlib
import shutil

from docker import DockerClient
from docker import APIClient as DockerAPIClient

from .exceptions import ROSWireException
from .description import SystemDescription, SystemDescriptionManager
from .system import System
from .proxy import ContainerProxy, ContainerProxyManager
from .definitions import FormatDatabase, PackageDatabase, TypeDatabase

logger: logging.Logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class ROSWire:
    """
    Provides an interface for building, analysing, and interacting with
    containerised ROS applications.

    Attributes
    ----------
    containers: ContainerProxyManager
        A manager for building and connecting to Docker containers.
    descriptions: SystemDescriptionManager
        A manager for building, loading, and storing static descriptions of
        ROS applications.
    workspace: str
        The absolute path of the workspace directory for this session. The
        workspace is used to store cache data and to store shared temporary
        directories between the host machine and containerised ROS
        applications.
    """
    def __init__(self,
                 dir_workspace: Optional[str] = None
                 ) -> None:
        if not dir_workspace:
            logger.debug("no workspace specified: using default workspace.")
            dir_home = os.path.expanduser("~")
            dir_workspace = os.path.join(dir_home, ".roswire")
            logger.debug("default workspace: %s", dir_workspace)
            if not os.path.exists(dir_workspace):
                logger.debug("initialising default workspace")
                os.mkdir(dir_workspace)
        else:
            logger.debug("using specified workspace: %s", dir_workspace)
            if not os.path.exists(dir_workspace):
                m = "workspace not found: {}".format(dir_workspace)
                raise ROSWireException(m)

        self.__dir_workspace = os.path.abspath(dir_workspace)
        self.__client_docker = DockerClient()
        self.__api_docker = \
            DockerAPIClient(base_url='unix://var/run/docker.sock')
        self.__containers = ContainerProxyManager(self.__client_docker,
                                                  self.__api_docker,
                                                  self.__dir_workspace)
        dir_descriptions = os.path.join(dir_workspace, 'descriptions')
        self.__descriptions = SystemDescriptionManager(self.__containers,
                                                       dir_descriptions)

    @property
    def workspace(self) -> str:
        return self.__dir_workspace

    @property
    def client_docker(self) -> DockerClient:
        return self.__client_docker

    @property
    def containers(self) -> ContainerProxyManager:
        return self.__containers

    @property
    def descriptions(self) -> SystemDescriptionManager:
        return self.__descriptions

    @contextlib.contextmanager
    def launch(self,
               image: str,
               description: Optional[SystemDescription] = None
               ) -> Iterator[System]:
        """Launches a ROS application using a provided Docker image.

        Parameters
        ----------
            image: str
                the name of the Docker image.
            description: Optional[SystemDescription]
                an optional static description of the ROS application.
                If no description is provided, ROSWire will attempt to load one
                from the cache or else build one.
        """
        if not description:
            description = self.descriptions.load_or_build(image)
        with self.containers.launch(image) as container:
            container = container
            yield System(container, description)



from typing import Iterator, Tuple
import contextlib
import logging
import time

import pytest

import roswire
import roswire.exceptions
from roswire import ROSWire, ROSProxy, System, SystemDescription
from roswire.proxy import ShellProxy, FileProxy, ContainerProxy

NAME_IMAGE = 'hello-world'


def test_description():
    rsw = ROSWire()
    desc = rsw.descriptions.build(NAME_IMAGE, save=False)
    assert set(desc.packages) == {
        'nodelet_core',
        'ros_core',
        'ros_comm',
        'ros_base',
        'ros_tutorials',
        'common_tutorials',
        'roscpp_core',
        'bond_core',
        'ros',
        'common_msgs',
        'actionlib',
        'actionlib_msgs',
        'actionlib_tutorials',
        'angles',
        'bond',
        'bondcpp',
        'bondpy',
        'catkin',
        'class_loader',
        'cmake_modules',
        'cpp_common',
        'diagnostic_msgs',
        'dynamic_reconfigure',
        'gencpp',
        'genlisp',
        'genmsg',
        'genpy',
        'geometry_msgs',
        'message_filters',
        'message_generation',
        'message_runtime',
        'mk',
        'nav_msgs',
        'nodelet',
        'nodelet_topic_tools',
        'nodelet_tutorial_math',
        'pluginlib',
        'pluginlib_tutorials',
        'rosbag',
        'rosbag_migration_rule',
        'rosbag_storage',
        'rosbash',
        'rosboost_cfg',
        'rosbuild',
        'rosclean',
        'rosconsole',
        'rosconsole_bridge',
        'roscpp',
        'roscpp_serialization',
        'roscpp_traits',
        'roscpp_tutorials',
        'roscreate',
        'rosgraph',
        'rosgraph_msgs',
        'roslang',
        'roslaunch',
        'roslib',
        'roslisp',
        'roslz4',
        'rosmake',
        'rosmaster',
        'rosmsg',
        'rosnode',
        'rosout',
        'rospack',
        'rosparam',
        'rospy',
        'rospy_tutorials',
        'rosservice',
        'rostest',
        'rostime',
        'rostopic',
        'rosunit',
        'roswtf',
        'sensor_msgs',
        'shape_msgs',
        'smclib',
        'std_msgs',
        'std_srvs',
        'stereo_msgs',
        'topic_tools',
        'trajectory_msgs',
        'turtle_actionlib',
        'turtlesim',
        'visualization_msgs',
        'xmlrpcpp'}



# -*- coding: utf-8 -*-
"""
This file provides data structures that represent ROS launch configurations.
"""
__all__ = ('ROSConfig', 'NodeConfig')

from typing import Tuple, FrozenSet, Optional, Dict, Any
import logging

import attr

from ...util import build_tuple
from ...exceptions import FailedToParseLaunchFile
from ...name import (namespace_join, canonical_name, name_is_global,
                     namespaces_of)

logger: logging.Logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@attr.s(frozen=True, slots=True)
class Parameter:
    name: str = attr.ib()
    value: str = attr.ib()  # TODO convert to appropriate type


@attr.s(frozen=True, slots=True)
class NodeConfig:
    namespace: str = attr.ib()
    name: str = attr.ib()
    typ: str = attr.ib()
    package: str = attr.ib()
    remappings: Tuple[Tuple[str, str], ...] = attr.ib(default=tuple())
    filename: Optional[str] = attr.ib(default=None)
    output: Optional[str] = attr.ib(default=None)
    required: bool = attr.ib(default=False)
    respawn: bool = attr.ib(default=False)
    respawn_delay: float = attr.ib(default=0.0)
    env_args: Tuple[Tuple[str, str], ...] = attr.ib(default=tuple())
    cwd: Optional[str] = attr.ib(default=None)
    args: Optional[str] = attr.ib(default=None)
    launch_prefix: Optional[str] = attr.ib(default=None)

    @property
    def full_name(self) -> str:
        return namespace_join(self.namespace, self.name)


@attr.s(frozen=True, slots=True)
class ROSConfig:
    nodes: FrozenSet[NodeConfig] = attr.ib(default=frozenset(),
                                           converter=frozenset)
    executables: Tuple[str, ...] = attr.ib(default=tuple())
    roslaunch_files: Tuple[str, ...] = attr.ib(default=tuple())
    params: Dict[str, Any] = attr.ib(factory=dict)
    clear_params: Tuple[str, ...] = attr.ib(default=tuple())
    errors: Tuple[str, ...] = attr.ib(default=tuple())

    def with_clear_param(self, ns: str) -> 'ROSConfig':
        """
        Specifies a parameter that should be cleared before new parameters
        are set.
        """
        ns = canonical_name(ns)
        if ns in self.clear_params:
            return self
        clear_params = self.clear_params + (ns,)
        return attr.evolve(self, clear_params=clear_params)

    def with_param(self, name: str, value: Any) -> 'ROSConfig':
        """Adds a parameter to this configuration."""
        params = self.params.copy()
        errors = self.errors

        if not name_is_global(name):
            m = f"expected parameter name to be global: {name}"
            raise FailedToParseLaunchFile(m)

        for parent_name in (n for n in namespaces_of(name) if n in params):
            err = f"parameter [{name}] conflicts with parent [{parent_name}]"
            errors = errors + (err,)

        params[name] = value
        return attr.evolve(self, params=params, errors=errors)

    def with_executable(self, executable: str) -> 'ROSConfig':
        """Specify an executable that should be run at launch."""
        executables = self.executables + (executable,)
        return attr.evolve(self, executables=executables)

    def with_roslaunch_file(self, filename: str) -> 'ROSConfig':
        roslaunch_files = self.roslaunch_files + (filename,)
        return attr.evolve(self, roslaunch_files=roslaunch_files)

    def with_node(self, node: NodeConfig) -> 'ROSConfig':
        logger.debug("adding node to config: %s", node)
        full_name = node.full_name
        used_names = {n.full_name for n in self.nodes}
        if node.full_name in used_names:
            m = 'multiple definitions of node [{}] in launch configuration'
            m = m.format(node.full_name)
            raise FailedToParseLaunchFile(m)
        nodes = self.nodes | frozenset({node})
        return attr.evolve(self, nodes=nodes)



from typing import Iterator
import contextlib
import tempfile
import shutil
import os

import pytest

from roswire.proxy import FileProxy

from test_basic import build_file_proxy


def test_exists():
    with build_file_proxy() as files:
        assert not files.exists('/foo')
        assert files.exists('/tmp')
        assert files.exists('/ros_ws/entrypoint.sh')
        assert not files.exists('/ros_ws/entrypoint.shhhhh')


def test_find():
    with build_file_proxy() as files:
        assert files.find('/', 'pkgs.rosinstall') == ['/ros_ws/pkgs.rosinstall']
        assert len(files.find('/ros_ws', 'package.xml')) == 154
        with pytest.raises(OSError):
            files.find('/dummy_dir', 'file')


def test_isfile():
    with build_file_proxy() as files:
        assert not files.isfile('/foo')
        assert not files.isfile('/tmp')
        assert files.isfile('/ros_ws/entrypoint.sh')


def test_isdir():
    with build_file_proxy() as files:
        assert not files.isdir('/foo')
        assert files.isdir('/tmp')
        assert files.isdir('/ros_ws')
        assert files.isdir('/ros_ws/')
        assert not files.isdir('/ros_ws/entrypoint.sh')
        assert not files.isdir('/ros_ws/entrypoint.sh/')


def test_islink():
    with build_file_proxy() as files:
        assert files.isfile('/ros_ws/src/ArduPilot/libraries/AP_HAL_F4Light/sbus.cpp')
        assert files.islink('/ros_ws/src/ArduPilot/libraries/AP_HAL_F4Light/sbus.cpp')
        assert files.isfile('/ros_ws/src/ArduPilot/libraries/AP_HAL_Linux/sbus.cpp')
        assert not files.islink('/ros_ws/src/ArduPilot/libraries/AP_HAL_Linux/sbus.cpp')


def test_listdir():
    with build_file_proxy() as files:
        assert files.listdir('/ros_ws') == [
            '.catkin_tools',
            'build',
            'devel',
            'entrypoint.sh',
            'logs',
            'pkgs.rosinstall',
            'src'
        ]

        # not a directory
        with pytest.raises(OSError):
            files.listdir('/ros_ws/pkgs.rosinstall')

        # not a file or directory
        with pytest.raises(OSError):
            files.listdir('/ros_ws/idontexist')

        # abs
        assert files.listdir('/ros_ws', absolute=True) == [
            '/ros_ws/.catkin_tools',
            '/ros_ws/build',
            '/ros_ws/devel',
            '/ros_ws/entrypoint.sh',
            '/ros_ws/logs',
            '/ros_ws/pkgs.rosinstall',
            '/ros_ws/src'
        ]


def test_mkdir():
    with build_file_proxy() as files:
        files.mkdir('/ros_ws/cool')
        assert files.isdir('/ros_ws/cool')
        assert 'cool' in files.listdir('/ros_ws')

        # directory already exists
        with pytest.raises(FileExistsError):
            files.mkdir('/ros_ws/cool')
        assert files.isdir('/ros_ws/cool')

        # intermediate directory doesn't exist
        with pytest.raises(FileNotFoundError):
            files.mkdir('/ros_ws/foo/bar')
        assert not files.isdir('/ros_ws/foo/bar')
        assert not files.isdir('/ros_ws/foo')

        # parent directory is a file
        with pytest.raises(NotADirectoryError):
            files.mkdir('/ros_ws/entrypoint.sh/foo')
        assert not files.exists('/ros_ws/entrypoint.sh/foo')
        assert not files.isdir('/ros_ws/entrypoint.sh')
        assert files.isfile('/ros_ws/entrypoint.sh')


def test_makedirs():
    with build_file_proxy() as files:
        # parent directory exists
        files.makedirs('/ros_ws/cool')
        assert files.isdir('/ros_ws/cool')

        # intermediate directory doesn't exist
        files.makedirs('/ros_ws/foo/bar')
        assert files.isdir('/ros_ws/foo/bar')

        # directory already exists
        with pytest.raises(FileExistsError):
            files.makedirs('/ros_ws/cool')
        assert files.isdir('/ros_ws/cool')
        files.makedirs('/ros_ws/cool', exist_ok=True)

        # path is a file
        with pytest.raises(FileExistsError):
            files.makedirs('/ros_ws/entrypoint.sh')
        with pytest.raises(FileExistsError):
            files.makedirs('/ros_ws/entrypoint.sh', exist_ok=True)

        # parent directory is a file
        with pytest.raises(NotADirectoryError):
            files.makedirs('/ros_ws/entrypoint.sh/foo')
        assert not files.exists('/ros_ws/entrypoint.sh/foo')
        assert not files.isdir('/ros_ws/entrypoint.sh')
        assert files.isfile('/ros_ws/entrypoint.sh')


def test_remove():
    with build_file_proxy() as files:
        assert files.isfile('/ros_ws/pkgs.rosinstall')
        files.remove('/ros_ws/pkgs.rosinstall')
        assert not files.exists('/ros_ws/pkgs.rosinstall')

        # remove non-existent file
        with pytest.raises(FileNotFoundError):
            files.remove('/foo/bar')

        # remove directory
        assert files.isdir('/ros_ws/build')
        with pytest.raises(IsADirectoryError):
            files.remove('/ros_ws/build')
        assert files.isdir('/ros_ws/build')


def test_rmdir():
    with build_file_proxy() as files:
        # create and remove an empty directory
        files.mkdir('/tmp/foo')
        assert 'foo' in files.listdir('/tmp')
        assert files.isdir('/tmp/foo')
        files.rmdir('/tmp/foo')
        assert not files.exists('/tmp/foo')

        # remove a file
        assert files.isfile('/ros_ws/pkgs.rosinstall')
        with pytest.raises(NotADirectoryError):
            files.rmdir('/ros_ws/pkgs.rosinstall')
        assert files.isfile('/ros_ws/pkgs.rosinstall')

        # remove a non-existent file/directory
        assert not files.exists('/tmp/foo')
        with pytest.raises(FileNotFoundError):
            files.rmdir('/tmp/foo')

        # remove a non-empty directory
        assert files.isdir('/ros_ws/src')
        with pytest.raises(OSError):
            files.rmdir('/ros_ws/src')
        assert files.isdir('/ros_ws/src')
        assert files.isdir('/ros_ws/src/ArduPilot')


def test_rmtree():
    with build_file_proxy() as files:
        # create and remove an empty directory
        files.mkdir('/tmp/foo')
        assert files.isdir('/tmp/foo')
        files.rmtree('/tmp/foo')
        assert not files.exists('/tmp/foo')

        # remove a file
        assert files.isfile('/ros_ws/pkgs.rosinstall')
        with pytest.raises(NotADirectoryError):
            files.rmdir('/ros_ws/pkgs.rosinstall')
        assert files.isfile('/ros_ws/pkgs.rosinstall')

        # remove a non-existent file/directory
        assert not files.exists('/tmp/foo')
        with pytest.raises(FileNotFoundError):
            files.rmdir('/tmp/foo')

        # remove a non-empty directory
        assert files.isdir('/ros_ws/src')
        files.rmtree('/ros_ws/src')
        assert not files.exists('/ros_ws/src')


def test_copy_to_host():
    with build_file_proxy() as files:
        # copy file
        expected = """
#!/bin/bash
set -e

# setup ros environment
source "/opt/ros/$ROS_DISTRO/setup.bash"
exec "$@"
""".lstrip()
        _, fn_host = tempfile.mkstemp()
        try:
            files.copy_to_host('/ros_entrypoint.sh', fn_host)
            with open(fn_host, 'r') as f:
                contents = f.read()
            assert contents == expected
        finally:
            os.remove(fn_host)

        # non-existent file
        with pytest.raises(FileNotFoundError):
            files.copy_to_host('/foo/bar', fn_host)

        # copy directory
        dir_host: str = tempfile.mkdtemp()
        dir_host_src: str = os.path.join(dir_host, 'src')
        try:
            files.copy_to_host('/ros_ws/src/geometry/eigen_conversions/src',
                               dir_host)
            assert os.path.isdir(dir_host_src)
            assert set(os.listdir(dir_host_src)) == {'eigen_kdl.cpp', 'eigen_msg.cpp'}
        finally:
            shutil.rmtree(dir_host)


def test_copy_from_host():
    with build_file_proxy() as files:
        # create temporary file on host
        expected = """
#!/bin/bash
set -e

# setup ros environment
source "/opt/ros/$ROS_DISTRO/setup.bash"
exec "$@"
""".lstrip()
        _, fn_host = tempfile.mkstemp()
        try:
            with open(fn_host, 'w') as f:
                f.write(expected)

            fn_container = '/tmp/foo'
            files.copy_from_host(fn_host, fn_container)
            assert files.isfile(fn_container)
            assert files.read(fn_container) == expected
            files.remove(fn_container)
        finally:
            os.remove(fn_host)

        # copy non-existent file
        assert not os.path.exists(fn_host)
        with pytest.raises(FileNotFoundError):
            files.copy_from_host(fn_host, '/tmp/bar')

        # copy directory
        dir_host = tempfile.mkdtemp()
        dir_container = '/tmp/cool'
        try:
            _, fn_host_foo = tempfile.mkstemp(dir=dir_host)
            fn_container_foo = os.path.join(dir_container,
                                            os.path.basename(fn_host_foo))
            with open(fn_host_foo, 'w') as f:
                f.write(expected)

            files.copy_from_host(dir_host, dir_container)
            assert files.isdir(dir_container)
            assert files.isfile(fn_container_foo)
            assert files.read(fn_container_foo) == expected
        finally:
            shutil.rmtree(dir_host)


def test_read():
    with build_file_proxy() as files:
        # read file
        expected = "geometry_msgs/TransformStamped[] transforms\n"
        assert files.read('/ros_ws/src/geometry/tf/msg/tfMessage.msg') == expected

        # read binary file
        binary = files.read('/ros_ws/src/geometry/tf/msg/tfMessage.msg', binary=True)
        text = binary.decode('utf-8')
        assert text == expected

        # non-existent file
        with pytest.raises(FileNotFoundError):
            files.read('/foo/bar')

        # directory
        with pytest.raises(IsADirectoryError):
            files.read('/ros_ws')


def test_write():
    with build_file_proxy() as files:
        # write to new file
        expected = "hello world"
        files.write('/tmp/foo', expected)
        assert files.isfile('/tmp/foo')
        assert files.read('/tmp/foo') == expected

        # overwrite existing file
        expected = "goodbye world"
        files.write('/tmp/foo', expected)
        assert files.read('/tmp/foo') == expected
        files.remove('/tmp/foo')

        # write to a file that belongs to a non-existent directory
        with pytest.raises(FileNotFoundError):
            files.write('/tmp/bar/bork', 'code things')


def test_mktemp():
    with build_file_proxy() as files:
        # create a temporary file
        fn = files.mktemp()
        assert files.isfile(fn)
        assert os.path.isabs(fn)  # bug #193
        assert fn.startswith('/tmp/')  # bug #193

        # use specified dir
        d = '/boop'
        files.makedirs(d, exist_ok=True)
        assert files.isdir(d)
        fn = files.mktemp(dirname=d)
        assert os.path.dirname(fn) == d
        assert files.isfile(fn)

        # use non-existent dir
        with pytest.raises(FileNotFoundError):
            d = '/idontexist'
            assert not files.isdir(d)
            files.mktemp(dirname=d)

        # add suffix
        fn = files.mktemp(suffix='.foo')
        assert fn.endswith('.foo')
        assert files.isfile(fn)

        # add prefix
        fn = files.mktemp(prefix='foo')
        assert os.path.basename(fn).startswith('foo')
        assert files.isfile(fn)


def test_tempfile():
    with build_file_proxy() as files:
        fn: str
        with files.tempfile() as fn:
            assert files.isfile(fn)
        assert not files.exists(fn)



import os

from roswire.bag.daikon import bag_to_decls

from test_bag import load_mavros_description

DIR_TEST = os.path.dirname(__file__)


def test_bag_to_decls():
    system_desc = load_mavros_description()
    fn_bag = os.path.join(DIR_TEST, 'example.bag')
    bag_to_decls(fn_bag, system_desc)



__all__ = ['ActionFormat']

from typing import Optional, List, Dict, Any
import os

import attr

from .msg import MsgFormat
from ..proxy import FileProxy
from .. import exceptions


@attr.s(frozen=True)
class ActionFormat:
    """
    Reference: http://wiki.ros.org/actionlib
    """
    package = attr.ib(type=str)
    name = attr.ib(type=str)
    definition = attr.ib(type=str)
    goal = attr.ib(type=MsgFormat)
    feedback = attr.ib(type=Optional[MsgFormat])
    result = attr.ib(type=Optional[MsgFormat])

    @staticmethod
    def from_file(package: str, fn: str, files: FileProxy) -> 'ActionFormat':
        """
        Constructs a message format from a .msg file for a given package.

        Parameters:
            package: the name of the package that provides the file.
            fn: the path to the .msg file.
            files: a proxy for accessing the filesystem.

        Raises:
            FileNotFoundError: if the given file cannot be found.
        """
        assert fn.endswith('.action'), \
            'action format files must end in .action'
        name: str = os.path.basename(fn[:-7])
        contents: str = files.read(fn)
        return ActionFormat.from_string(package, name, contents)

    @staticmethod
    def from_string(package: str, name: str, s: str) -> 'ActionFormat':
        """
        Constructs an action format from its description.

        Raises:
            ParsingError: if the description cannot be parsed.
        """
        goal: MsgFormat
        feed: Optional[MsgFormat]
        res: Optional[MsgFormat]

        name_goal = f"{name}Goal"
        name_feed = f"{name}Feedback"
        name_res = f"{name}Result"

        sections: List[str] = [ss.strip() for ss in s.split('\n---')]
        try:
            s_goal, s_res, s_feed = sections
        except ValueError:
            m = "failed to parse action description: expected three sections."
            raise exceptions.ParsingError(m)

        goal = MsgFormat.from_string(package, name_goal, s_goal)

        if s_feed:
            feed = MsgFormat.from_string(package, name_feed, s_feed)
        else:
            feed = None

        if s_res:
            res = MsgFormat.from_string(package, name_res, s_res)
        else:
            res = None

        return ActionFormat(package, name, s, goal, feed, res)

    @staticmethod
    def from_dict(d: Dict[str, Any],
                  *,
                  package: Optional[str] = None
                  ) -> 'ActionFormat':
        name: str = d['name']
        definition: str = d['definition']
        if package is None:
            assert d['package'] is not None
            package = d['package']

        res: Optional[MsgFormat] = None
        feed: Optional[MsgFormat] = None
        goal: MsgFormat = \
            MsgFormat.from_dict(d['goal'], package=package, name=f'{name}Goal')

        if 'result' in d:
            res = MsgFormat.from_dict(d['result'],
                                      package=package,
                                      name=f'{name}Result')
        if 'feedback' in d:
            feed = MsgFormat.from_dict(d['feedback'],
                                       package=package,
                                       name=f'{name}Feedback')

        return ActionFormat(package, name, definition, goal, feed, res)

    def to_dict(self) -> Dict[str, Any]:
        d = {'package': self.package,
             'name': self.name,
             'definition': self.definition,
             'goal': self.goal.to_dict()}
        if self.feedback:
            d['feedback'] = self.feedback.to_dict()
        if self.result:
            d['result'] = self.result.to_dict()
        return d

    @property
    def fullname(self) -> str:
        return f"{self.package}/{self.name}"



__all__ = ('Package', 'PackageDatabase')

from typing import (Tuple, List, Dict, Union, Any, Iterator, Collection,
                    Mapping)
import os

import attr
import shlex

from .msg import MsgFormat
from .srv import SrvFormat
from .action import ActionFormat
from ..proxy import FileProxy, ShellProxy


@attr.s(frozen=True)
class Package:
    name: str = attr.ib()
    path: str = attr.ib()
    messages: Tuple[MsgFormat, ...] = attr.ib(converter=tuple)
    services: Tuple[SrvFormat, ...] = attr.ib(converter=tuple)
    actions: Tuple[ActionFormat, ...] = attr.ib(converter=tuple)

    @staticmethod
    def build(path: str, files: FileProxy) -> 'Package':
        """
        Constructs a description of a package located at a given path.
        """
        name: str = os.path.basename(path)
        messages: List[MsgFormat] = []
        services: List[SrvFormat] = []
        actions: List[ActionFormat] = []

        if not files.isdir(path):
            raise FileNotFoundError(f"directory does not exist: {path}")

        dir_msg = os.path.join(path, 'msg')
        dir_srv = os.path.join(path, 'srv')
        dir_action = os.path.join(path, 'action')

        if files.isdir(dir_msg):
            messages = [MsgFormat.from_file(name, f, files)
                        for f in files.listdir(dir_msg, absolute=True)
                        if f.endswith('.msg')]
        if files.isdir(dir_srv):
            services = [SrvFormat.from_file(name, f, files)
                        for f in files.listdir(dir_srv, absolute=True)
                        if f.endswith('.srv')]
        if files.isdir(dir_action):
            actions = [ActionFormat.from_file(name, f, files)
                       for f in files.listdir(dir_action, absolute=True)
                       if f.endswith('.action')]

        return Package(name,  # type: ignore
                       path,
                       messages,
                       services,
                       actions)

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> 'Package':
        name: str = d['name']
        messages: List[MsgFormat] = [MsgFormat.from_dict(dd, package=name)
                                     for dd in d.get('messages', [])]
        services: List[SrvFormat] = [SrvFormat.from_dict(dd, package=name)
                                     for dd in d.get('services', [])]
        actions: List[ActionFormat] = [ActionFormat.from_dict(dd, package=name)
                                       for dd in d.get('actions', [])]
        return Package(d['name'],  # type: ignore
                       d['path'],
                       messages,
                       services,
                       actions)

    def to_dict(self) -> Dict[str, Any]:
        d = {'name': self.name,
             'path': self.path,
             'messages': [m.to_dict() for m in self.messages],
             'services': [s.to_dict() for s in self.services],
             'actions': [a.to_dict() for a in self.actions]}
        return d


class PackageDatabase(Mapping[str, Package]):
    @staticmethod
    def paths(shell: ShellProxy, files: FileProxy) -> List[str]:
        """
        Parses the contents of the ROS_PACKAGE_PATH environment variable for a
        given shell.
        """
        path_str = shell.environ('ROS_PACKAGE_PATH')
        package_paths: List[str] = path_str.strip().split(':')
        paths: List[str] = []
        for path in package_paths:
            try:
                all_packages = files.find(path, 'package.xml')
            except OSError:
                # path is not a directory
                continue
            package_dirs = [os.path.dirname(p) for p in all_packages]
            paths.extend(package_dirs)
        return paths

    @staticmethod
    def from_paths(files: FileProxy,
                   paths: List[str],
                   ignore_bad_paths: bool = True
                   ) -> 'PackageDatabase':
        """
        Constructs a package database from a list of the paths of the packages
        belonging to the database.

        Parameters:
            files: access to the filesystem.
            paths: a list of the absolute paths of the packages.
        """
        packages: List[Package] = []
        for p in paths:
            try:
                package = Package.build(p, files)
            except FileNotFoundError:
                if not ignore_bad_paths:
                    raise
            packages.append(package)
        return PackageDatabase(packages)

    def __init__(self, packages: Collection[Package]) -> None:
        self.__contents = {p.name: p for p in packages}

    def __len__(self) -> int:
        """
        Returns a count of the number of packages within this database.
        """
        return len(self.__contents)

    def __getitem__(self, name: str) -> Package:
        """
        Fetches the description for a given package.

        Raises:
            KeyError: if no package exists with the given name.
        """
        return self.__contents[name]

    def __iter__(self) -> Iterator[str]:
        """
        Returns an iterator over the names of the packages contained within
        this database.
        """
        yield from self.__contents

    @staticmethod
    def from_dict(d: List[Dict[str, Any]]) -> 'PackageDatabase':
        return PackageDatabase([Package.from_dict(dd) for dd in d])

    def to_dict(self) -> List[Dict[str, Any]]:
        return [p.to_dict() for p in self.values()]



__all__ = ('Constant', 'ConstantValue', 'Field', 'MsgFormat', 'Message')

from typing import (Type, Optional, Any, Union, Tuple, List, Dict, ClassVar,
                    Collection, Set, Iterator, Mapping, Callable, BinaryIO)
from io import BytesIO
import logging
import functools
import hashlib
import struct
import re
import os

import attr
from toposort import toposort_flatten as toposort

from .base import is_builtin, Time, Duration
from .decode import is_simple
from ..proxy import FileProxy
from .. import exceptions

logger: logging.Logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

R_TYPE = r"[a-zA-Z0-9_/]+(?:\[\d*\])?"
R_NAME = r"[a-zA-Z0-9_/]+"
R_VAL = r".+"
R_COMMENT = r"(#.*)?"
R_FIELD = re.compile(f"^\s*({R_TYPE})\s+({R_NAME})\s*{R_COMMENT}$")
R_STRING_CONSTANT = re.compile(f"^\s*string\s+(\w+)\s*=\s*(.+)\s*$")
R_OTHER_CONSTANT = re.compile(f"^\s*(\w+)\s+(\w+)\s*=\s*([^\s]+).*$")
R_BLANK = re.compile(f"^\s*{R_COMMENT}$")

ConstantValue = Union[str, int, float]


@attr.s(frozen=True, str=False)
class Constant:
    typ = attr.ib(type=str)
    name = attr.ib(type=str)
    value = attr.ib(type=Union[str, int, float])

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> 'Constant':
        return Constant(d['type'], d['name'], d['value'])

    def to_dict(self) -> Dict[str, Any]:
        return {'type': self.typ,
                'name': self.name,
                'value': self.value}

    def __str__(self) -> str:
        return f"{self.typ} {self.name}={str(self.value)}"


@attr.s(frozen=True, str=False)
class Field:
    typ: str = attr.ib()
    name: str = attr.ib()

    @property
    def is_array(self) -> bool:
        return '[' in self.typ

    @property
    def is_simple(self) -> bool:
        return not self.is_array and is_simple(self.typ)

    @property
    def length(self) -> Optional[int]:
        if not self.is_array:
            return None
        sz = self.typ.partition('[')[2].partition(']')[0]
        if sz == '':
            return None
        return int(sz)

    @property
    def base_type(self) -> str:
        return self.typ.partition('[')[0] if self.is_array else self.typ

    @property
    def base_typ(self) -> str:
        return self.base_type

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> 'Field':
        return Field(d['type'], d['name'])

    def to_dict(self) -> Dict[str, str]:
        return {'type': self.typ,
                'name': self.name}

    def without_package_name(self) -> 'Field':
        typ = self.typ.partition('/')[2] if '/' in self.typ else self.typ
        return Field(typ, self.name)

    def __str__(self) -> str:
        return f"{self.typ} {self.name}"


@attr.s(frozen=True)
class MsgFormat:
    package: str = attr.ib()
    name: str = attr.ib()
    definition: str = attr.ib()
    fields: Tuple[Field, ...] = attr.ib(converter=tuple)
    constants: Tuple[Constant, ...] = attr.ib(converter=tuple)

    @staticmethod
    def toposort(fmts: Collection['MsgFormat']) -> List['MsgFormat']:
        fn_to_fmt: Dict[str, MsgFormat] = {f.fullname: f for f in fmts}
        fn_to_deps: Dict[str, Set[str]] = \
            {fn: {f.base_typ for f in fmt.fields if not is_builtin(f.base_typ)}
             for fn, fmt in fn_to_fmt.items()}
        toposorted = list(toposort(fn_to_deps))
        return [fn_to_fmt[fn] for fn in toposorted]

    @staticmethod
    def from_file(package: str, fn: str, files: FileProxy) -> 'MsgFormat':
        """
        Constructs a message format from a .msg file for a given package.

        Parameters:
            package: the name of the package that provides the file.
            fn: the path to the .msg file.
            files: a proxy for accessing the filesystem.

        Raises:
            FileNotFoundError: if the given file cannot be found.
        """
        assert fn.endswith('.msg'), 'message format files must end in .msg'
        name: str = os.path.basename(fn[:-4])
        contents: str = files.read(fn)
        return MsgFormat.from_string(package, name, contents)

    @staticmethod
    def from_string(package: str, name: str, text: str) -> 'MsgFormat':
        """
        Constructs a message format from its description.

        Raises:
            ParsingError: if the description cannot be parsed.
        """
        typ: str
        name_const: str
        fields: List[Field] = []
        constants: List[Constant] = []

        for line in text.split('\n'):
            m_blank = R_BLANK.match(line)
            m_string_constant = R_STRING_CONSTANT.match(line)
            m_other_constant = R_OTHER_CONSTANT.match(line)
            m_field = R_FIELD.match(line)

            if m_blank:
                continue
            elif m_string_constant:
                name_const, val = m_string_constant.group(1, 2)
                constant = Constant('string', name_const, val)
                constants.append(constant)
            elif m_other_constant:
                typ, name_const, val_str = m_other_constant.group(1, 2, 3)
                val = val_str  # FIXME convert value
                constant = Constant(typ, name_const, val)
                constants.append(constant)
            elif m_field:
                typ, name_field = m_field.group(1, 2)

                # resolve the type of the field
                typ_resolved = typ
                base_typ = typ.partition('[')[0]
                if typ == 'Header':
                    typ_resolved = 'std_msgs/Header'
                elif '/' not in typ and not is_builtin(base_typ):
                    typ_resolved = f'{package}/{typ}'

                if typ != typ_resolved:
                    logger.debug("resolved type [%s]: %s", typ, typ_resolved)
                    typ = typ_resolved

                field: Field = Field(typ, name_field)
                fields.append(field)
            else:
                raise exceptions.ParsingError(f"failed to parse line: {line}")

        return MsgFormat(package, name, text, fields, constants)  # type: ignore  # noqa

    @staticmethod
    def from_dict(d: Dict[str, Any],
                  *,
                  package: Optional[str] = None,
                  name: Optional[str] = None
                  ) -> 'MsgFormat':
        if not package:
            package = d['package']
        if not name:
            name = d['name']
        definition = d['definition']
        fields = [Field.from_dict(dd) for dd in d.get('fields', [])]
        constants = [Constant.from_dict(dd) for dd in d.get('constants', [])]
        return MsgFormat(package, name, definition, fields, constants)  # type: ignore  # noqa

    def to_dict(self) -> Dict[str, Any]:
        d: Dict[str, Any] = {'package': self.package,
                             'name': self.name,
                             'definition': self.definition}
        if self.fields:
            d['fields'] = [f.to_dict() for f in self.fields]
        if self.constants:
            d['constants'] = [c.to_dict() for c in self.constants]
        return d

    @property
    def fullname(self) -> str:
        return f"{self.package}/{self.name}"

    def flatten(self,
                name_to_format: Mapping[str, 'MsgFormat'],
                ctx: Tuple[str, ...] = tuple()
                ) -> Iterator[Tuple[Tuple[str, ...], Field]]:
        for field in self.fields:
            if field.is_array or is_builtin(field.typ):
                yield (ctx, field)
            else:
                fmt = name_to_format[field.typ]
                yield from fmt.flatten(name_to_format, ctx + (field.name,))

    def md5text(self, name_to_msg: Mapping[str, 'MsgFormat']) -> str:
        """Computes the MD5 text for this format."""
        lines: List[str] = []
        lines += [str(c) for c in self.constants]
        for f in self.fields:
            if is_builtin(f.base_type):
                lines += [str(f.without_package_name())]
            else:
                f_md5 = name_to_msg[f.base_type].md5sum(name_to_msg)
                lines += [f'{f_md5} {f.name}']
        return '\n'.join(lines)

    def md5sum(self, name_to_msg: Mapping[str, 'MsgFormat']) -> str:
        """Computes the MD5 sum for this format."""
        logger.debug("generating md5sum: %s", self.fullname)
        txt = self.md5text(name_to_msg)
        logger.debug("generated md5 text [%s]:\n%s", self.fullname, txt)
        md5sum = hashlib.md5(txt.encode('utf-8')).hexdigest()
        logger.debug("generated md5sum [%s]: %s", self.fullname, md5sum)
        return md5sum


class Message:
    """Base class used by all messages."""
    format: ClassVar[MsgFormat]

    @staticmethod
    def _to_dict_value(val: Any) -> Any:
        typ = type(val)

        if typ in (Time, Duration) or isinstance(typ, Message):
            return val.to_dict()

        if typ in (list, tuple):
            if not typ:
                return []
            typ_item = type(typ[0])
            if typ_item == Time or isinstance(typ_item, Message):
                return [vv.to_dict() for vv in val]
            # includes (str, int, float)
            return list(val)

        # includes (str, int, float)
        return val

    def to_dict(self) -> Dict[str, Any]:
        d: Dict[str, Any] = {}
        for field in self.format.fields:
            name: str = field.name
            val = getattr(self, field.name)
            d[name] = self._to_dict_value(val)
        return d

    @classmethod
    def md5sum(cls) -> str:
        """Returns the md5sum for this message type."""
        raise NotImplementedError

    @classmethod
    def read(cls, b: BinaryIO) -> 'Message':
        raise NotImplementedError

    @classmethod
    def decode(cls, b: bytes) -> 'Message':
        return cls.read(BytesIO(b))

    def write(self, b: BinaryIO) -> None:
        """Writes a binary encoding of this message to a given stream."""
        raise NotImplementedError

    def encode(self) -> bytes:
        """Returns a binary encoding of this message."""
        b = BytesIO()
        self.write(b)
        return b.getvalue()



# -*- coding: utf-8 -*-
__all__ = ('System',)

from typing import Iterator
from uuid import UUID
import contextlib
import logging

from .description import SystemDescription
from .definitions import TypeDatabase, FormatDatabase, PackageDatabase
from .proxy import (ShellProxy, ROSProxy, FileProxy, ContainerProxy,
                    CatkinProxy, CatkinToolsProxy, CatkinMakeProxy)

logger: logging.Logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class System:
    def __init__(self,
                 container: ContainerProxy,
                 description: SystemDescription
                 ) -> None:
        self.__container = container
        self.__description = description

    @property
    def container(self) -> ContainerProxy:
        return self.__container

    @property
    def description(self) -> SystemDescription:
        return self.__description

    @property
    def uuid(self) -> UUID:
        return self.__container.uuid

    @property
    def ws_host(self) -> str:
        return self.__container.ws_host

    @property
    def ip_address(self) -> str:
        return self.__container.ip_address

    @property
    def shell(self) -> ShellProxy:
        return self.__container.shell

    def catkin(self, directory: str) -> CatkinProxy:
        """Returns an interface to a given catkin workspace."""
        # TODO decide whether to use catkin_tools or catkin_make
        return self.catkin_tools(directory)

    def catkin_tools(self, directory: str) -> CatkinToolsProxy:
        """Returns an interface to a given catkin tools workspace."""
        return CatkinToolsProxy(self.shell, directory)

    def catkin_make(self, directory: str) -> CatkinMakeProxy:
        """Returns an interface to a given catkin_make workspace."""
        return CatkinMakeProxy(self.shell, directory)

    @property
    def files(self) -> FileProxy:
        return self.__container.files

    @contextlib.contextmanager
    def roscore(self, port: int = 11311) -> Iterator[ROSProxy]:
        assert port > 1023
        cmd = "roscore -p {}".format(port)
        self.shell.non_blocking_execute(cmd)
        try:
            yield ROSProxy(description=self.description,
                           shell=self.shell,
                           files=self.files,
                           ws_host=self.ws_host,
                           ip_address=self.ip_address,
                           port=port)
        finally:
            self.shell.execute("pkill roscore")



from typing import Type

import pytest

from roswire.definitions.base import Time, Duration


def test_duration_between():
    sec_to_nsec = 1000000000
    t_start = Time(150, 300)
    t_stop = Time(150, 350)
    assert Duration.between(t_start, t_stop) == Duration(0, 50)
    assert Duration.between(t_start, t_start) == Duration(0, 0)
    with pytest.raises(AssertionError):
        Duration.between(t_stop, t_start)

    t_start = Time(150, 300)
    t_stop = Time(200, 300)
    assert Duration.between(t_start, t_stop) == Duration(50, 0)

    t_start = Time(150, 300)
    t_stop = Time(200, 50)
    assert Duration.between(t_start, t_stop) == Duration(49, sec_to_nsec - 250)

    t_start = Time(150, 50)
    t_stop = Time(200, 300)
    assert Duration.between(t_start, t_stop) == Duration(50, 250)


def check_timelike(t: Type):
    assert t(100, 250) < t(100, 300)
    assert t(100, 250) <= t(100, 300)
    assert not (t(100, 300) < t(100, 250))
    assert t(100, 300) > t(100, 250)
    assert t(100, 300) >= t(100, 250)


    assert t(100, 250) <= t(100, 300)
    assert t(500, 0) > t(0, 500)

    assert min(t(0, 50), t(100, 200), t(300, 500)) == t(0, 50)
    assert max(t(0, 50), t(100, 200), t(300, 500)) == t(300, 500)
    assert min(t(500, 0), t(100, 200), t(300, 500)) == t(100, 200)


def test_time_compare():
    check_timelike(Time)


def test_duration_compare():
    check_timelike(Duration)



# -*- coding: utf-8 -*-
"""
This file implements a proxy for parsing the contents of launch files.
"""
__all__ = ('LaunchFileReader',)

from typing import (List, Optional, Sequence, Collection, Dict, Any, Mapping,
                    Tuple, Union, overload)
from copy import deepcopy
import logging
import xml.etree.ElementTree as ET

import attr
import yaml

from .rosparam import load_from_yaml_string as load_rosparam_from_string
from .config import ROSConfig, NodeConfig, Parameter
from .context import LaunchContext
from ..substitution import resolve as resolve_args
from ..shell import ShellProxy
from ..file import FileProxy
from ...name import (namespace_join, global_name, namespace, name_is_global,
                     name_is_private)
from ...exceptions import FailedToParseLaunchFile

logger: logging.Logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

_TAG_TO_LOADER = {}


def _read_contents(tag: ET.Element) -> str:
    """Reads the text contents of an XML element."""
    # FIXME add support for CDATA -- possibly via lxml or xml.dom?
    return ''.join(t.text for t in tag if t.text)


def _parse_bool(attr: str, val: str) -> bool:
    """Parses a boolean value from an XML attribute."""
    val = val.lower()
    if val == 'true':
        return True
    if val == 'false':
        return False

    m = f'illegal boolean attribute [{attr}]: {val}'
    raise FailedToParseLaunchFile(m)


def _parse_float(attr: str, val: str) -> float:
    """Parses a float value from an XML attribute."""
    if not val:
        m = f'empty string used by float attribute [{attr}]'
        raise FailedToParseLaunchFile(m)
    try:
        return float(val)
    except ValueError:
        m = f'failed to parse attribute [{attr}] to float: {val}'
        raise FailedToParseLaunchFile(m)


def convert_str_to_type(s: str, typ: str) -> Union[bool, int, str, float]:
    if typ == 'auto':
        if s.lower() in ('true', 'false'):
            return convert_str_to_type(s, 'bool')
        if s.isnumeric() and '.' in s:
            return convert_str_to_type(s, 'float')
        if s.isnumeric():
            return convert_str_to_type(s, 'int')
        return convert_str_to_type(s, 'str')

    if typ in ('str', 'string'):
        return s
    if typ in ('float', 'double'):
        return float(s)
    if typ == 'int':
        return int(s)
    if typ in ('bool', 'boolean'):
        s = s.lower()
        if s in ('true', '1'):
            return True
        if s in ('false', '0'):
            return False
        raise ValueError(f"illegal 'bool' value: {s}")

    raise ValueError(f'unknown parameter type: {typ}')


def tag(name: str, legal_attributes: Collection[str] = tuple()):
    legal_attributes = frozenset(legal_attributes)

    def wrap(loader):
        def wrapped(self,
                    ctx: LaunchContext,
                    cfg: ROSConfig,
                    elem: ET.Element
                    ) -> Tuple[LaunchContext, ROSConfig]:
            logger.debug("parsing <%s> tag", name)
            for attribute in elem.attrib:
                if attribute not in legal_attributes:
                    m = '<{}> tag contains illegal attribute: {}'
                    raise FailedToParseLaunchFile(m.format(name, attribute))
            ctx, cfg = loader(self, ctx, cfg, elem)
            return ctx, cfg
        _TAG_TO_LOADER[name] = wrapped
        return wrapped

    return wrap


class LaunchFileReader:
    def __init__(self, shell: ShellProxy, files: FileProxy) -> None:
        self.__shell = shell
        self.__files = files

    def _parse_file(self, fn: str) -> ET.Element:
        """Parses a given XML launch file to a root XML element."""
        root = ET.fromstring(self.__files.read(fn))
        if root.tag != 'launch':
            m = 'root of launch file must have <launch></launch> tags'
            raise FailedToParseLaunchFile(m)
        return root

    def _load_tags(self,
                   ctx: LaunchContext,
                   cfg: ROSConfig,
                   tags: Sequence[ET.Element]
                   ) -> Tuple[LaunchContext, ROSConfig]:
        for tag in (t for t in tags if t.tag in _TAG_TO_LOADER):
            loader = _TAG_TO_LOADER[tag.tag]
            ctx, cfg = loader(self, ctx, cfg, tag)
        return ctx, cfg

    @tag('param', ['name', 'value', 'type', 'textfile', 'binfile', 'command'])
    def _load_param_tag(self,
                        ctx: LaunchContext,
                        cfg: ROSConfig,
                        tag: ET.Element
                        ) -> Tuple[LaunchContext, ROSConfig]:
        name = self._read_required(tag, 'name', ctx)
        typ = self._read_optional(tag, 'type', ctx) or 'auto'
        logger.debug("adding parameter [%s] with type [%s]", name, typ)

        # obtain value from either 'value', 'textfile', 'binfile', or
        # 'command' exclusively.
        val = self._read_optional(tag, 'value', ctx)
        textfile = self._read_optional(tag, 'textfile', ctx)
        binfile = self._read_optional(tag, 'binfile', ctx)
        command = self._read_optional(tag, 'command', ctx)
        value_candidates = [val, textfile, binfile, command]

        if len([c for c in value_candidates if c is not None]) != 1:
            m = '<param> must have exactly one value/textfile/binfile/command'
            raise FailedToParseLaunchFile(m)

        value: Any
        if val is not None:
            value = convert_str_to_type(val, typ)
        if textfile is not None:
            value = self.__files.read(textfile)
        if binfile is not None:
            value = self.__files.read(binfile, binary=True)
        if command is not None:
            retcode, value, duration = self.__shell.execute(command)

        logger.debug("obtained value for parameter [%s]: %s", name, value)

        # compute the fully qualified name
        if name_is_global(name):
            fullname = name
        elif name_is_private(name):
            fullname = namespace_join(ctx.namespace, name[1:])
        else:
            fullname = namespace_join(ctx.namespace, name)

        # register the parameter
        cfg = cfg.with_param(fullname, value)

        return ctx, cfg

    @tag('rosparam', ['command', 'ns', 'file', 'param', 'subst_value'])
    def _load_rosparam_tag(self,
                           ctx: LaunchContext,
                           cfg: ROSConfig,
                           tag: ET.Element
                           ) -> Tuple[LaunchContext, ROSConfig]:
        filename = self._read_optional(tag, 'file', ctx)
        subst_value = self._read_optional_bool(tag, 'subst_value', ctx, False)
        ns = self._read_optional(tag, 'ns', ctx) or ''
        param = self._read_optional(tag, 'param', ctx) or ''
        param = namespace_join(ns, param)
        full_param = namespace_join(ctx.namespace, param)
        value = _read_contents(tag)

        cmd: str = self._read_optional(tag, 'command', ctx) or 'load'
        if cmd not in ('load', 'delete', 'dump'):
            m = f"<rosparam> unsupported 'command': {cmd}"
            raise FailedToParseLaunchFile(m)

        if cmd == 'load' and not filename:
            m = "<rosparam> load command requires 'filename' attribute"
            raise FailedToParseLaunchFile(m)

        if cmd == 'load':
            assert filename is not None  # mypy can't work this out
            if not self.__files.isfile(filename):
                m = f"<rosparam> file does not exist: {filename}"
                raise FailedToParseLaunchFile(m)

        if cmd == 'delete' and filename is not None:
            m = "<rosparam> command:delete does not support filename"
            raise FailedToParseLaunchFile(m)

        # handle load command
        if cmd == 'load':
            assert filename is not None  # mypy can't work this out
            yml_text = self.__files.read(filename)
            if subst_value:
                yml_text = self._resolve_args(yml_text, ctx)
            logger.debug("parsing rosparam YAML:\n%s", yml_text)
            data = load_rosparam_from_string(yml_text)
            logger.debug("rosparam values: %s", data)
            if not isinstance(data, dict) and not param:
                m = "<rosparam> requires 'param' for non-dictionary values"
                raise FailedToParseLaunchFile(m)
            cfg = cfg.with_param(full_param, data)

        # handle dump command
        if cmd == 'dump':
            m = "'dump' command is currently not supported in <rosparam>"
            raise NotImplementedError(m)

        # handle delete command
        if cmd == 'delete':
            m = "'delete' command is currently not supported in <rosparam>"
            raise NotImplementedError(m)

        return ctx, cfg

    @tag('remap', ['from', 'to'])
    def _load_remap_tag(self,
                        ctx: LaunchContext,
                        cfg: ROSConfig,
                        tag: ET.Element
                        ) -> Tuple[LaunchContext, ROSConfig]:
        frm = self._read_required(tag, 'from', ctx)
        to = self._read_required(tag, 'to', ctx)
        ctx = ctx.with_remapping(frm, to)
        return ctx, cfg

    @tag('node', ['name', 'type', 'pkg', 'required', 'clear_params',
                  'respawn', 'namespace', 'output', 'args'])
    def _load_node_tag(self,
                       ctx: LaunchContext,
                       cfg: ROSConfig,
                       tag: ET.Element
                       ) -> Tuple[LaunchContext, ROSConfig]:
        name = self._read_required(tag, 'name', ctx)
        package = self._read_required(tag, 'pkg', ctx)
        node_type = self._read_required(tag, 'type', ctx)
        output = self._read_optional(tag, 'output', ctx)
        launch_prefix = self._read_optional(tag, 'launch-prefix', ctx)
        cwd = self._read_optional(tag, 'cwd', ctx)
        args = self._read_optional(tag, 'args', ctx)
        required = self._read_optional_bool(tag, 'required', ctx, False)
        respawn = self._read_optional_bool(tag, 'respawn', ctx, False)
        respawn_delay = \
            self._read_optional_float(tag, 'respawn_delay', ctx, 0.0)

        # create context
        ns = self._read_namespace(ctx, tag)
        ctx_child = ctx.node_child(ns, name)

        # handle 'clear_params'
        if self._read_optional_bool(tag, 'clear_params', ctx, False):
            clear_ns = global_name(ctx_child.namespace)
            cfg = cfg.with_clear_param(clear_ns)

        # handle nested tags
        allowed = {'env', 'remap', 'param', 'rosparam'}
        nested_tags = [t for t in tag if t.tag in allowed]
        ctx_child, cfg = self._load_tags(ctx_child, cfg, nested_tags)

        node = NodeConfig(name=name,
                          namespace=namespace(ctx_child.namespace),
                          package=package,
                          cwd=cwd,
                          args=args,
                          required=required,
                          respawn=respawn,
                          respawn_delay=respawn_delay,
                          output=output,
                          remappings=ctx_child.remappings,
                          launch_prefix=launch_prefix,
                          filename=ctx.filename,
                          env_args=ctx_child.env_args,
                          typ=node_type)
        cfg = cfg.with_node(node)
        return ctx, cfg

    @tag('arg', ['name', 'default', 'value', 'doc'])
    def _load_arg_tag(self,
                      ctx: LaunchContext,
                      cfg: ROSConfig,
                      tag: ET.Element
                      ) -> Tuple[LaunchContext, ROSConfig]:
        name = self._read_required(tag, 'name', ctx)
        value = self._read_optional(tag, 'value', ctx)
        default = self._read_optional(tag, 'default', ctx)
        doc = self._read_optional(tag, 'doc', ctx)
        ctx = ctx.with_arg(name=name,
                           value=value,
                           default=default,
                           doc=doc)
        return ctx, cfg

    @tag('env', ['name', 'value'])
    def _load_env_tag(self,
                      ctx: LaunchContext,
                      cfg: ROSConfig,
                      tag: ET.Element
                      ) -> Tuple[LaunchContext, ROSConfig]:
        name = self._read_required(tag, 'name', ctx)
        value = self._read_required(tag, 'value', ctx)
        ctx = ctx.with_env_arg(name, value)
        return ctx, cfg

    @tag('include', ['file', 'pass_all_args', 'ns', 'clear_params'])
    def _load_include_tag(self,
                          ctx: LaunchContext,
                          cfg: ROSConfig,
                          tag: ET.Element
                          ) -> Tuple[LaunchContext, ROSConfig]:
        include_filename = self._read_required(tag, 'file', ctx)
        logger.debug("include file: %s", include_filename)
        cfg = cfg.with_roslaunch_file(include_filename)

        # create context
        ns = self._read_namespace(ctx, tag)
        ctx_child = ctx.include_child(ns, include_filename)

        # handle 'clear_params'
        if self._read_optional_bool(tag, 'clear_params', ctx, False):
            if not ns:
                m = "'ns' must be specified to use 'clear_params'"
                raise FailedToParseLaunchFile(m)
            cfg = cfg.with_clear_param(ctx_child.namespace)

        # if instructed to pass along args, then those args must be added to
        # the child context
        if self._read_optional_bool(tag, 'pass_all_args', ctx, False):
            ctx_child = ctx_child.with_pass_all_args()

        # handle child tags
        child_tags = [t for t in tag if t.tag in ('env', 'arg')]
        ctx_child, cfg = self._load_tags(ctx_child, cfg, child_tags)
        ctx_child = ctx_child.process_include_args()
        logger.debug("prepared include context: %s", ctx_child)

        logger.debug("loading include file")
        launch = self._parse_file(include_filename)
        ctx_child, cfg = self._load_tags(ctx_child, cfg, list(launch))

        return ctx, cfg

    def _read_namespace(self,
                        ctx: LaunchContext,
                        tag: ET.Element
                        ) -> Optional[str]:
        """Fetches the value of the optional 'namespace' attribute."""
        ns = self._read_optional(tag, 'namespace', ctx)
        if ns == '':
            m = f"<{tag.tag}> has empty attribute [namespace]"
            raise FailedToParseLaunchFile(m)
        return ns

    @overload
    def _read_optional_bool(self,
                            elem: ET.Element,
                            attrib: str,
                            ctx: LaunchContext,
                            default: None
                            ) -> Optional[bool]: ...

    @overload
    def _read_optional_bool(self,
                            elem: ET.Element,
                            attrib: str,
                            ctx: LaunchContext,
                            default: bool
                            ) -> bool: ...

    def _read_optional_bool(self, elem, attrib, ctx, default=None):
        s = self._read_optional(elem, attrib, ctx)
        if s is None:
            return default
        return _parse_bool(attrib, s)

    @overload
    def _read_optional_float(self,
                             elem: ET.Element,
                             attrib: str,
                             ctx: LaunchContext,
                             default: None
                             ) -> Optional[float]: ...

    @overload
    def _read_optional_float(self,
                             elem: ET.Element,
                             attrib: str,
                             ctx: LaunchContext,
                             default: float
                             ) -> float: ...

    def _read_optional_float(self, elem, attrib, ctx, default=None):
        s = self._read_optional(elem, attrib, ctx)
        if s is None:
            return default
        return _parse_float(attrib, s)

    def _read_optional(self,
                       elem: ET.Element,
                       attrib: str,
                       ctx: LaunchContext
                       ) -> Optional[str]:
        """Reads the string value of an optional attribute of an element."""
        if attrib not in elem.attrib:
            return None
        return self._read_required(elem, attrib, ctx)

    def _read_required(self,
                       elem: ET.Element,
                       attrib: str,
                       ctx: LaunchContext
                       ) -> str:
        """Reads the string value of a required attribute of an element."""
        return self._resolve_args(elem.attrib[attrib], ctx)

    def _resolve_args(self, s: str, ctx: LaunchContext) -> str:
        """Resolves all substitution args in a given string."""
        logger.debug("resolve [%s] with context: %s", s, ctx.resolve_dict)
        return resolve_args(self.__shell, self.__files, s, ctx.resolve_dict)

    def read(self, fn: str, argv: Optional[Sequence[str]] = None) -> ROSConfig:
        """Parses the contents of a given launch file.

        Returns
        -------
        ROSConfig
            A description of the launch configuration.

        Reference
        ---------
            http://wiki.ros.org/roslaunch/XML/node
            http://docs.ros.org/kinetic/api/roslaunch/html/roslaunch.xmlloader.XmlLoader-class.html
        """
        cfg = ROSConfig()
        ctx = LaunchContext(namespace='/', filename=fn)
        if argv:
            ctx = ctx.with_argv(argv)

        launch = self._parse_file(fn)
        ctx, cfg = self._load_tags(ctx, cfg, list(launch))
        logger.debug("launch configuration: %s", cfg)
        return cfg



import subprocess
import time
import os

import pytest

from roswire.proxy import ShellProxy
from roswire.exceptions import EnvNotFoundError

from test_basic import build_shell_proxy


def test_wait():
    with build_shell_proxy() as shell:
        p = shell.popen('sleep 30 && echo "hello world"')
        with pytest.raises(subprocess.TimeoutExpired):
            p.wait(1)
        assert p.wait(35) == 0

        p = shell.popen('sleep 10 && exit 1')
        assert p.wait() == 1


def test_local_to_host_pid():
    with build_shell_proxy() as shell:
        p = shell.popen('sleep 60')
        pid_host = shell.local_to_host_pid(p.pid)
        assert pid_host is not None


def test_environ():
    with build_shell_proxy() as shell:
        assert shell.environ('ROS_PACKAGE_PATH')
        with pytest.raises(EnvNotFoundError):
            shell.environ('FOOBARFOOBARFOO')



import pytest

from roswire.proxy import CatkinProxy

from test_basic import build_hello_world


def test_clean_and_build():
    with build_hello_world() as (sut, ros):
        catkin = sut.catkin('/ros_ws')
        catkin.clean()
        catkin.build()



# -*- coding: utf-8 -*-
__all__ = ('SrvFormat',)

from typing import Optional, List, Dict, Any
import os

import attr

from .msg import MsgFormat, Constant, Field
from ..proxy import FileProxy
from .. import exceptions


@attr.s(frozen=True)
class SrvFormat:
    package = attr.ib(type=str)
    name = attr.ib(type=str)
    definition = attr.ib(type=str)
    request = attr.ib(type=Optional[MsgFormat])
    response = attr.ib(type=Optional[MsgFormat])

    @staticmethod
    def from_file(package: str, fn: str, files: FileProxy) -> 'SrvFormat':
        """Constructs a service format from a .srv file for a given package.

        Parameters:
            package: the name of the package that provides the file.
            fn: the path to the .srv file.
            files: a proxy for accessing the filesystem.

        Raises:
            FileNotFoundError: if the given file cannot be found.
        """
        assert fn.endswith('.srv'), 'service format files must end in .srv'
        name: str = os.path.basename(fn[:-4])
        contents: str = files.read(fn)
        return SrvFormat.from_string(package, name, contents)

    @staticmethod
    def from_string(package: str, name: str, s: str) -> 'SrvFormat':
        """Constructs a service format from its description.

        Raises:
            ParsingError: if the description cannot be parsed.
        """
        req: Optional[MsgFormat] = None
        res: Optional[MsgFormat] = None
        name_req = f"{name}Request"
        name_res = f"{name}Response"

        sections: List[str] = [ss.strip() for ss in s.split('---')]
        assert len(sections) < 3
        s_req = sections[0]
        s_res = sections[1] if len(sections) > 1 else ''

        if s_req:
            req = MsgFormat.from_string(package, name_req, s_req)
        if s_res:
            res = MsgFormat.from_string(package, name_res, s_res)

        return SrvFormat(package, name, s, req, res)  # type: ignore

    @staticmethod
    def from_dict(d: Dict[str, Any],
                  *,
                  package: Optional[str] = None
                  ) -> 'SrvFormat':
        req: Optional[MsgFormat] = None
        res: Optional[MsgFormat] = None
        name: str = d['name']
        definition: str = d['definition']
        if package is None:
            assert d['package'] is not None
            package = d['package']

        if 'request' in d:
            req = MsgFormat.from_dict(d['request'],
                                      package=package,
                                      name=f'{name}Request')
        if 'response' in d:
            res = MsgFormat.from_dict(d['response'],
                                      package=package,
                                      name=f'{name}Response')

        return SrvFormat(package, name, definition, req, res)

    def to_dict(self) -> Dict[str, Any]:
        d: Dict[str, Any] = {'package': self.package,
                             'name': self.name,
                             'definition': self.definition}
        if self.request:
            d['request'] = self.request.to_dict()
        if self.response:
            d['response'] = self.response.to_dict()
        return d

    @property
    def fullname(self) -> str:
        return f"{self.package}/{self.name}"



__all__ = ('ShellProxy',)

from typing import Tuple, Optional, Dict, Any, Iterator, List
from timeit import default_timer as timer
from subprocess import TimeoutExpired
import os
import shlex
import logging
import threading
import time
import signal

import psutil
from docker import DockerClient
from docker import APIClient as DockerAPIClient
from docker.models.containers import Container as DockerContainer

from ..util import Stopwatch
from .. import exceptions

logger: logging.Logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class Popen:
    def __init__(self,
                 args: str,
                 uid: str,
                 shell: 'ShellProxy',
                 api_docker: DockerAPIClient,
                 exec_id: int,
                 stream: Iterator[bytes]
                 ) -> None:
        self.__args = args
        self.__uid = uid
        self.__shell = shell
        self.__api_docker = api_docker
        self.__exec_id = exec_id
        self.__stream = stream
        self.__pid: Optional[int] = None
        self.__returncode: Optional[int] = None

    def _inspect(self) -> Dict[str, Any]:
        return self.__api_docker.exec_inspect(self.__exec_id)

    @property
    def stream(self) -> Iterator[str]:
        for line_bytes in self.__stream:
            yield line_bytes.decode('utf-8')

    @property
    def args(self) -> str:
        return self.__args

    def __uid_to_pid(self) -> Optional[int]:
        prefix = f'/bin/bash -c echo {self.__uid} > /dev/null && '
        cmd = f'ps -eo pid,cmd | grep "{prefix}"'
        while not self.finished:
            code, output, duration = self.__shell.execute(cmd)
            for line in output.split('\n'):
                pid_str, _, p_cmd = line.strip().partition(' ')
                if p_cmd.startswith(prefix):
                    return int(pid_str)
        return None

    @property
    def pid(self) -> Optional[int]:
        if not self.__pid and not self.finished:
            self.__pid = self.__uid_to_pid()
        return self.__pid

    @property
    def finished(self) -> bool:
        """True if the process has exited; False if not."""
        return self.returncode is not None

    @property
    def returncode(self) -> Optional[int]:
        if self.__returncode is None:
            self.__returncode = self._inspect()['ExitCode']
        return self.__returncode

    def send_signal(self, sig: int) -> None:
        pid = self.pid
        if pid:
            self.__shell.send_signal(pid, sig)

    def kill(self) -> None:
        """Kills the process via a SIGKILL signal."""
        self.send_signal(signal.SIGKILL)

    def terminate(self) -> None:
        """Terminates the process via a SIGTERM signal."""
        self.send_signal(signal.SIGTERM)

    def poll(self) -> Optional[int]:
        """Checks if the process has terminated and returns its returncode."""
        return self.returncode

    def wait(self, time_limit: Optional[float] = None) -> int:
        """Blocks until the process has terminated.

        Parameters
        ----------
        time_limit: Optional[float] = None
            An optional time limit.

        Raises
        ------
        subprocess.TimeoutExpired:
            if the process does not terminate within the specified timeout.
        """
        stopwatch = Stopwatch()
        stopwatch.start()
        while not self.finished:
            if time_limit and stopwatch.duration > time_limit:
                raise TimeoutExpired(self.args, time_limit)
            time.sleep(0.05)
        assert self.returncode is not None
        return self.returncode


class ShellProxy:
    """Provides shell access for a given Docker container."""
    def __init__(self,
                 api_docker: DockerAPIClient,
                 container_docker: DockerContainer,
                 container_pid: int
                 ) -> None:
        self.__api_docker = api_docker
        self.__container_docker = container_docker
        self.__container_pid = container_pid

    def exec_id_to_host_pid(self, exec_id: str) -> int:
        """Returns the host PID for a given exec command."""
        return self.__api_docker.exec_inspect(exec_id)['Pid']

    def local_to_host_pid(self, pid_local: int) -> Optional[int]:
        """Finds the host PID for a process inside this shell."""
        ctr_pids = [self.__container_pid]
        info = self.__api_docker.inspect_container(self.__container_docker.id)
        ctr_pids += [self.exec_id_to_host_pid(i) for i in info['ExecIDs']]

        # obtain a list of all processes inside this container
        ctr_procs: List[psutil.Process] = []
        for pid in ctr_pids:
            proc = psutil.Process(pid)
            ctr_procs.append(proc)
            ctr_procs += proc.children(recursive=True)

        # read /proc/PID/status to find the namespace mapping
        for proc in ctr_procs:
            fn_proc = f'/proc/{proc.pid}/status'
            with open(fn_proc, 'r') as fh_proc:
                lines = filter(lambda l: l.startswith('NSpid'),
                               fh_proc.readlines())
                for line in lines:
                    proc_host_pid, proc_local_pid = \
                        [int(p) for p in line.strip().split('\t')[1:3]]
                    if proc_local_pid == pid_local:
                        return proc_host_pid

        return None

    def send_signal(self, pid: int, sig: int) -> None:
        self.execute(f'kill -{sig} {pid}', user='root')

    def __generate_popen_uid(self, command: str) -> str:
        id_thread = threading.get_ident()
        h = abs(hash((id_thread, timer(), command)))
        return str(h)

    def environ(self, var: str) -> str:
        """Reads the value of a given environment variable inside this shell.

        Raises
        ------
        EnvNotFoundError
            if no environment variable exists with the given name.
        """
        cmd = f'test -v {var} && echo "${{{var}}}"'
        retcode, val, duration = self.execute(cmd)
        if retcode != 0:
            raise exceptions.EnvNotFoundError(var)
        return val

    def instrument(self,
                   command: str,
                   time_limit: Optional[int] = None,
                   kill_after: int = 1,
                   identifier: Optional[str] = None
                   ) -> str:
        logger.debug("instrumenting command: %s", command)
        q = shlex.quote
        command = f'source /.environment && {command}'
        if identifier:
            command = f'echo {q(identifier)} > /dev/null && {command}'
        command = f'/bin/bash -c {q(command)}'
        if time_limit:
            command = (f'timeout --kill-after={kill_after} '
                       f'--signal=SIGTERM {time_limit} {command}')
        logger.debug("instrumented command: %s", command)
        return command

    def popen(self,
              command: str,
              *,
              stdout: bool = True,
              stderr: bool = True,
              user: Optional[str] = None,
              context: Optional[str] = None,
              time_limit: Optional[int] = None,
              kill_after: int = 1
              ) -> Popen:
        uid_popen = self.__generate_popen_uid(command)
        id_container = self.__container_docker.id
        api_docker = self.__api_docker
        command_orig = command
        command = self.instrument(command, time_limit, kill_after,
                                  identifier=uid_popen)
        exec_resp = api_docker.exec_create(id_container, command,
                                           tty=True,
                                           stdout=stdout,
                                           stderr=stderr)
        exec_id = exec_resp['Id']
        stream = api_docker.exec_start(exec_id, stream=True)
        return Popen(command_orig,
                     uid_popen,
                     self,
                     api_docker,
                     exec_id,
                     stream)

    def execute(self,
                command: str,
                *,
                stdout: bool = True,
                stderr: bool = True,
                user: Optional[str] = None,
                context: Optional[str] = None,
                time_limit: Optional[int] = None,
                kill_after: int = 1
                ) -> Tuple[int, str, float]:
        logger.debug("executing command: %s", command)
        dockerc = self.__container_docker
        command = self.instrument(command, time_limit, kill_after)

        timer = Stopwatch()
        timer.start()
        # https://docker-py.readthedocs.io/en/stable/containers.html
        retcode, output = \
            dockerc.exec_run(command, stdout=stdout, stderr=stderr, workdir=context)  # noqa
        timer.stop()
        duration = timer.duration
        output = output.decode('utf-8').rstrip('\n')
        logger.debug("executed command [%s] (retcode: %d; time: %.3f s)\n%s",
                     command, retcode, timer.duration, output)
        return retcode, output, timer.duration

    def non_blocking_execute(self,
                             *args,
                             **kwargs
                             ) -> None:
        thread = threading.Thread(target=ShellProxy.execute,
                                  args=(self, *args),
                                  kwargs=kwargs)
        thread.start()



# -*- coding: utf-8 -*-
class ROSWireException(Exception):
    """Base class used by all ROSWire exceptions."""


class FailedToParseLaunchFile(ROSWireException):
    """An attempt to parse a launch file failed."""


class EnvNotFoundError(ROSWireException):
    """A given environment variable could not be found."""
    def __init__(self, var: str) -> None:
        m = f"could not find enviroment variable: {var}"
        super().__init__(m)


class SubstitutionError(ROSWireException):
    """An error occurred during substitution argument handling."""


class CatkinException(ROSWireException):
    """Base class used by all Catkin-related exceptions."""


class CatkinBuildFailed(CatkinException):
    """The attempt to build via catkin failed."""
    def __init__(self, retcode: int, reason: str) -> None:
        msg = f"catkin build failed [retcode: {retcode}]: {reason}"
        super().__init__(msg)


class CatkinCleanFailed(CatkinException):
    """An attempt to clean the catkin workspace failed."""
    def __init__(self, retcode: int, reason: str) -> None:
        msg = f"catkin clean failed [retcode: {retcode}]: {reason}"
        super().__init__(msg)


class PlayerNotStarted(ROSWireException):
    """Playback has not begun."""


class PlayerAlreadyStarted(ROSWireException):
    """Playback has already started."""


class PlayerAlreadyStopped(ROSWireException):
    """Playback has already stopped."""


class PlayerFailure(ROSWireException):
    """An unexpected error occurred during playback."""
    def __init__(self, retcode: int, stdout: str) -> None:
        self.retcode = retcode
        self.stdout = stdout
        msg = "unexpected failure during bag playback [return code: {}]:\n{}"
        msg = msg.format(retcode, stdout)
        super().__init__(msg)


class PlayerTimeout(ROSWireException):
    """Playback did not complete within the specified timeout."""


class RecorderAlreadyStarted(ROSWireException):
    """Recording has already started."""


class RecorderNotStarted(ROSWireException):
    """Recording has not begun."""


class RecorderAlreadyStopped(ROSWireException):
    """Recording has already stopped."""


class ParsingError(ROSWireException):
    """ROSWire failed to parse a given file/string."""


class NodeNotFoundError(KeyError, ROSWireException):
    """No node was found with the given name."""
    def __init__(self, name: str) -> None:
        super().__init__(f"node not found: {name}")


class ServiceNotFoundError(KeyError, ROSWireException):
    """No service was found with the given name."""
    def __init__(self, name: str) -> None:
        super().__init__(f"service not found: {name}")


class ParameterNotFoundError(KeyError, ROSWireException):
    """No parameter was found with the given name."""
    def __init__(self, name: str) -> None:
        super().__init__(f"parameter not found: {name}")



from typing import Iterator, Tuple
import os
import contextlib
import logging
import time

import pytest

import roswire
import roswire.exceptions
from roswire import ROSWire, ROSProxy, System, SystemDescription
from roswire.proxy import ShellProxy, FileProxy, ContainerProxy
from roswire.description import SystemDescription
from roswire.definitions import TypeDatabase, FormatDatabase, PackageDatabase

DIR_TEST = os.path.dirname(__file__)


def load_hello_world_type_db() -> TypeDatabase:
    fn_db_format = os.path.join(DIR_TEST,
                                'format-databases/helloworld.formats.yml')
    db_format = FormatDatabase.load(fn_db_format)
    return TypeDatabase.build(db_format)


def load_hello_world_description() -> SystemDescription:
    fn_db_format = os.path.join(DIR_TEST,
                                'format-databases/helloworld.formats.yml')
    db_format = FormatDatabase.load(fn_db_format)
    db_type = TypeDatabase.build(db_format)
    desc = SystemDescription(sha256='foo',
                             types=db_type,
                             formats=db_format,
                             packages=PackageDatabase([]))
    return desc


@contextlib.contextmanager
def build_test_environment() -> Iterator[Tuple[System, ROSProxy]]:
    rsw = ROSWire()
    image = 'brass'
    desc = SystemDescription(image, [], [], [])
    with rsw.launch(image, desc) as sut:
        with sut.roscore() as ros:
            time.sleep(5)
            yield (sut, ros)


@contextlib.contextmanager
def build_ardu() -> Iterator[Tuple[System, ROSProxy]]:
    rsw = ROSWire()
    with rsw.launch('brass') as sut:
        with sut.roscore() as ros:
            time.sleep(5)
            yield (sut, ros)


@contextlib.contextmanager
def build_hello_world() -> Iterator[Tuple[System, ROSProxy]]:
    rsw = ROSWire()
    image = 'roswire/helloworld:buggy'
    desc = load_hello_world_description()
    with rsw.launch(image, desc) as sut:
        with sut.roscore() as ros:
            time.sleep(5)
            yield (sut, ros)


@contextlib.contextmanager
def build_shell_proxy() -> Iterator[ShellProxy]:
    rsw = ROSWire()
    image = 'brass'
    desc = SystemDescription(image, [], [], [])
    with rsw.launch(image, desc) as sut:
        yield sut.shell


@contextlib.contextmanager
def build_file_proxy() -> Iterator[FileProxy]:
    rsw = ROSWire()
    image = 'brass'
    desc = SystemDescription(image, [], [], [])
    with rsw.launch(image, desc) as sut:
        yield sut.files


@contextlib.contextmanager
def build_file_and_shell_proxy() -> Iterator[Tuple[FileProxy, ShellProxy]]:
    rsw = ROSWire()
    image = 'brass'
    desc = SystemDescription(image, [], [], [])
    with rsw.launch(image, desc) as sut:
        yield sut.files, sut.shell


def test_parameters():
    with build_test_environment() as (sut, ros):
        assert ros.topic_to_type == {'/rosout': 'rosgraph_msgs/Log',
                                     '/rosout_agg': 'rosgraph_msgs/Log'}

        assert '/rosversion' in ros.parameters
        assert '/rosdistro' in ros.parameters

        assert '/hello' not in ros.parameters
        ros.parameters['/hello'] = 'world'
        assert '/hello' in ros.parameters
        assert ros.parameters['/hello'] == 'world'

        del ros.parameters['/hello']
        assert 'hello' not in ros.parameters
        with pytest.raises(KeyError):
            ros.parameters['/hello']


def test_arducopter():
    logging.basicConfig()
    with build_ardu() as (sut, ros):
        db_type = sut.description.types
        db_fmt = sut.description.formats
        cmd = ' '.join([
            "/ros_ws/src/ArduPilot/build/sitl/bin/arducopter",
            "--model copter"
        ])
        sut.shell.non_blocking_execute(cmd)

        ros.launch('mavros', 'apm.launch', fcu_url='tcp://127.0.0.1:5760@5760')
        time.sleep(10)


        print(list(ros.nodes))
        assert set(ros.nodes) == {'/mavros', '/rosout'}
        assert '/mavros' in ros.nodes
        assert '/rosout' in ros.nodes
        assert '/cool' not in ros.nodes

        # with ros.record('/tmp/baggio.bag') as recorder:
        #     time.sleep(30)

        with pytest.raises(rozzy.exceptions.NodeNotFoundError):
            ros.nodes['/cool']

        node_mavros = ros.nodes['/mavros']
        assert node_mavros.name == '/mavros'
        assert node_mavros.pid > 0
        print(f"URL: {node_mavros.url}")
        print(f"PID: {node_mavros.pid}")

        print(list(ros.services))
        print(ros.services['/mavros/set_mode'])
        assert '/mavros/set_mode' in ros.services

        with pytest.raises(rozzy.exceptions.ServiceNotFoundError):
            ros.services['/coolio']

        assert '/coolio' not in ros.services

        # arm the copter
        req: Message = db_type.from_dict(db_fmt.messages['mavros_msgs/SetModeRequest'], {
            'base_mode': 64,
            'custom_mode': ''
        })
        res = ros.services['/mavros/set_mode'].call(req)
        assert res.success


if __name__ == '__main__':
    test_arducopter()



# -*- coding: utf-8 -*-
"""
This module is used to build descriptions of containerised ROS applications,
and provides a manager for persisting those descriptions to disk.
"""
__all__ = ('SystemDescription', 'SystemDescriptionManager')

from typing import Union, Dict, Any
from uuid import UUID
import contextlib
import logging
import os

import yaml
import attr
from docker.models.images import Image as DockerImage

from .definitions import TypeDatabase, FormatDatabase, PackageDatabase
from .proxy import ContainerProxyManager

logger: logging.Logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@attr.s(slots=True)
class SystemDescription:
    """
    A description of the packages, types, and specifications within
    a containerised ROS application.

    Attributes
    ----------
        sha256: str
            The ID of the Docker image for the application.
        types: TypeDatabase
            A database of types for the application.
        formats: FormatDatabase
            A database of message, service and action specifications.
        packages:
            A database of the packages contained within the application.
    """
    sha256: str = attr.ib()
    types: TypeDatabase = attr.ib()
    formats: FormatDatabase = attr.ib()
    packages: PackageDatabase = attr.ib()

    @staticmethod
    def from_file(fn: str) -> 'SystemDescription':
        """Loads a description from a given file."""
        with open(fn, 'r') as f:
            d = yaml.safe_load(f)
        return SystemDescription.from_dict(d)

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> 'SystemDescription':
        """Constructs a description from a JSON dictionary."""
        sha256: str = d['sha256']
        packages = PackageDatabase.from_dict(d['packages'])
        formats = FormatDatabase.build(packages)
        types = TypeDatabase.build(formats)
        return SystemDescription(sha256=sha256,
                                 packages=packages,
                                 formats=formats,
                                 types=types)

    def to_dict(self) -> Dict[str, Any]:
        """Produces a JSON dictionary for this description."""
        return {'sha256': self.sha256,
                'packages': self.packages.to_dict()}


class SystemDescriptionManager:
    """
    Builds and stores static descriptions of containerised ROS applications.

    Attributes
    ----------
    dir: str
        The absolute path of the directory where descriptions are stored.
    """
    def __init__(self,
                 containers: ContainerProxyManager,
                 dir_cache: str
                 ) -> None:
        self.__containers = containers

        # ensure that the cache directory exists
        self.__dir_cache = dir_cache
        os.makedirs(dir_cache, exist_ok=True)

    @property
    def dir(self) -> str:
        return self.__dir_cache

    def load(self,
             image_or_tag: Union[str, DockerImage]
             ) -> SystemDescription:
        """
        Attempts to load the description for a given Docker image from disk.

        Parameters
        ----------
        image_or_tag: Union[str, DockerImage]
            the name or object for the Docker image.

        Returns
        -------
        SystemDescription
            A description of the application contained within the given image.

        Raises
        ------
        FileNotFoundError:
            if no description for the given image is stored on disk.
        """
        sha256 = self.__containers.image_sha256(image_or_tag)
        fn = os.path.join(self.__dir_cache, sha256)
        try:
            return SystemDescription.from_file(fn)
        except FileNotFoundError:
            logger.exception("failed to load description for image: %s",
                             image_or_tag)
            raise

    def saved(self, image_or_tag: Union[str, DockerImage]) -> bool:
        """
        Determines whether a description for a given image has been saved.
        """
        sha256 = self.__containers.image_sha256(image_or_tag)
        fn = os.path.join(self.__dir_cache, sha256)
        return os.path.exists(fn)

    def save(self, description: SystemDescription) -> None:
        """Saves a given description to disk."""
        fn = os.path.join(self.__dir_cache, description.sha256)
        yml = description.to_dict()
        with open(fn, 'w') as f:
            yaml.dump(yml, f, default_flow_style=False)

    def build(self,
              image_or_tag: Union[str, DockerImage],
              save: bool = True
              ) -> SystemDescription:
        """
        Builds a description of the ROS application contained within a given
        image and optionally saves that description to disk.


        Returns
        -------
        SystemDescription
            A description of the application contained within the given image.
        """
        sha256 = self.__containers.image_sha256(image_or_tag)
        with self.__containers.launch(image_or_tag) as container:
            paths = PackageDatabase.paths(container.shell, container.files)
            db_package = PackageDatabase.from_paths(container.files, paths)
        db_format = FormatDatabase.build(db_package)
        db_type = TypeDatabase.build(db_format)

        description = SystemDescription(sha256=sha256,
                                        packages=db_package,
                                        formats=db_format,
                                        types=db_type)
        if save:
            self.save(description)
        return description

    def load_or_build(self,
                      image_or_tag: Union[str, DockerImage],
                      save: bool = True
                      ) -> SystemDescription:
        """
        Attempts to load a description for a given image from disk, and if
        none can be found, a description will be built from scratch and
        optionally saved.

        Parameters
        ----------
        image_or_tag: Union[str, DockerImage]
            the name or object for the Docker image.
        save: bool
            if :code:`True`, the description will be saved to disk in the
            event that it needs to be built from scratch.

        Returns
        -------
        SystemDescription
            A description of the application contained within the given image.
        """
        if self.saved(image_or_tag):
            return self.load(image_or_tag)
        return self.build(image_or_tag, save)



import pytest

from roswire.definitions import (TypeDatabase, FormatDatabase, MsgFormat,
                                 Time, Duration)

from test_bag import load_mavros_type_db


def test_encode_and_decode():
    db_type = load_mavros_type_db()
    Header = db_type['std_msgs/Header']
    orig = Header(seq=32, stamp=Time(secs=9781, nsecs=321), frame_id='')
    assert orig == Header.decode(orig.encode())



# -*- coding: utf-8 -*-
__all__ = ('CatkinProxy', 'CatkinToolsProxy', 'CatkinMakeProxy')

from typing import Optional, List
import abc
import shlex
import logging

from .shell import ShellProxy
from ..exceptions import CatkinBuildFailed, CatkinCleanFailed

logger: logging.Logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class CatkinProxy(abc.ABC):
    def __init__(self, shell: ShellProxy, directory: str) -> None:
        """Constructs a catkin proxy for a given workspace."""
        self._directory = directory
        self._shell = shell

    @property
    def directory(self) -> str:
        """The directory of this catkin workspace."""
        return self._directory

    @abc.abstractmethod
    def clean(self,
              packages: Optional[List[str]] = None,
              orphans: bool = False,
              context: Optional[str] = None
              ) -> None:
        """Cleans all products of the catkin workspace.

        Raises
        ------
        CatkinCleanFailed:
            if the attempt to clean resulted in a non-zero return code.
        """
        ...

    @abc.abstractmethod
    def build(self,
              packages: Optional[List[str]] = None,
              no_deps: bool = False,
              pre_clean: bool = False,
              jobs: Optional[int] = None,
              cmake_args: Optional[List[str]] = None,
              make_args: Optional[List[str]] = None,
              context: Optional[str] = None,
              time_limit: Optional[int] = None
              ) -> None:
        """Attempts to build a catkin workspace.

        Raises
        ------
        CatkinBuildFailed:
            if the attempt to build resulted in a non-zero return code.
        """
        ...


class CatkinToolsProxy(CatkinProxy):
    """Provides an interface to a catkin workspace created via catkin tools."""
    def clean(self,
              packages: Optional[List[str]] = None,
              orphans: bool = False,
              context: Optional[str] = None
              ) -> None:
        shell = self._shell
        command = ['catkin', 'clean', '-y']
        if orphans:
            command += ['--orphans']
        if packages:
            command += [shlex.quote(p) for p in packages]
        if not context:
            context = self.directory

        command_str = ' '.join(command)
        logger.debug("cleaning via: %s", command_str)
        retcode, output, duration_secs = \
            shell.execute(command_str, context=context)
        duration_mins = duration_secs / 60
        logger.debug("clean completed after %.2f minutes [retcode: %d]:\n%s",
                     duration_mins, retcode, output)

        if retcode != 0:
            raise CatkinCleanFailed(retcode, output)

    def build(self,
              packages: Optional[List[str]] = None,
              no_deps: bool = False,
              pre_clean: bool = False,
              jobs: Optional[int] = None,
              cmake_args: Optional[List[str]] = None,
              make_args: Optional[List[str]] = None,
              context: Optional[str] = None,
              time_limit: Optional[int] = None
              ) -> None:
        shell = self._shell
        command = ['catkin', 'build', '--no-status', '--no-notify']
        if packages:
            command += [shlex.quote(p) for p in packages]
        if no_deps:
            command += ['--no-deps']
        if pre_clean:
            command += ['--pre-clean']
        if cmake_args:
            command += cmake_args
        if make_args:
            command += make_args
        if not context:
            context = self.directory

        command_str = ' '.join(command)
        logger.debug("building via: %s", command_str)
        retcode, output, duration_secs = \
            shell.execute(command_str, context=context, time_limit=time_limit)
        duration_mins = duration_secs / 60
        logger.debug("build completed after %.2f minutes [retcode: %d]:\n%s",
                     duration_mins, retcode, output)

        if retcode != 0:
            raise CatkinBuildFailed(retcode, output)


class CatkinMakeProxy(CatkinProxy):
    """Provides an interface to a catkin workspace created via catkin_make."""
    def clean(self,
              packages: Optional[List[str]] = None,
              orphans: bool = False,
              context: Optional[str] = None
              ) -> None:
        raise NotImplementedError

    def build(self,
              packages: Optional[List[str]] = None,
              no_deps: bool = False,
              pre_clean: bool = False,
              jobs: Optional[int] = None,
              cmake_args: Optional[List[str]] = None,
              make_args: Optional[List[str]] = None,
              context: Optional[str] = None,
              time_limit: Optional[int] = None
              ) -> None:
        raise NotImplementedError



# -*- coding: utf-8 -*-
"""
This module provides handling of XML substitution arguments, which are used
by XML launch and xacro files.
"""
__all__ = ('resolve',)

from typing import Optional, Dict, Any
import os
import re
import shlex
import functools
import logging

from .shell import ShellProxy
from .file import FileProxy
from ..exceptions import EnvNotFoundError, SubstitutionError

logger: logging.Logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

R_ARG = re.compile(r'\$\(.+?\)')


def resolve_arg(shell: ShellProxy,
                files: FileProxy,
                s: str,
                context: Optional[Dict[str, Any]] = None
                ) -> str:
    """
    Raises
    ------
    EnvNotFoundError
        if a given environment variable is not found.
    """
    if not context:
        context = {}

    logger.debug("resolving substitution argument: %s", s)
    s = s[2:-1]
    logger.debug("stripped delimiters: %s", s)
    kind, *params = s.split(' ')
    logger.debug("argument kind: %s", kind)

    if kind == 'env':
        return shell.environ(params[0])
    if kind == 'optenv':
        try:
            return shell.environ(params[0])
        except EnvNotFoundError:
            return ' '.join(params[1:])
    if kind == 'dirname':
        try:
            dirname = os.path.dirname(context['filename'])
        except KeyError:
            m = 'filename is not provided by the launch context'
            raise SubstitutionError(m)
        dirname = os.path.normpath(dirname)
        return dirname
    if kind == 'arg':
        arg_name = params[0]
        if 'arg' not in context or arg_name not in context['arg']:
            m = f'arg not supplied to launch context [{arg_name}]'
            raise SubstitutionError(m)
        return context['arg'][arg_name]
    if kind == 'find':
        package = params[0]
        cmd = f'rospack find {shlex.quote(package)}'
        retcode, location, duration = shell.execute(cmd)
        if retcode != 0:
            raise SubstitutionError(f'failed to locate package: {package}')
        return location.strip()

    # TODO $(anon name)
    return s


def resolve(shell: ShellProxy,
            files: FileProxy,
            s: str,
            context: Optional[Dict[str, str]] = None
            ) -> str:
    # TODO $(eval ...)
    if s.startswith('$(eval ') and s[-1] == ')':
        raise NotImplementedError
    r = functools.partial(resolve_arg, shell, files)
    return R_ARG.sub(lambda m: r(m.group(0), context), s)



import os
import logging
import tempfile

import yaml
import pytest

from roswire import ROSWire
from roswire.bag import BagReader, BagWriter
from roswire.description import SystemDescription
from roswire.definitions import TypeDatabase, FormatDatabase, PackageDatabase
from roswire.definitions import Time, Duration

from test_basic import build_ardu, build_hello_world

DIR_TEST = os.path.dirname(__file__)

ms_to_ns = 10 ** 6


def load_mavros_type_db() -> TypeDatabase:
    fn_db_format = os.path.join(DIR_TEST,
                                'format-databases/mavros.formats.yml')
    db_format = FormatDatabase.load(fn_db_format)
    return TypeDatabase.build(db_format)


def load_mavros_description() -> SystemDescription:
    fn_db_format = os.path.join(DIR_TEST,
                                'format-databases/mavros.formats.yml')
    db_format = FormatDatabase.load(fn_db_format)
    db_type = TypeDatabase.build(db_format)
    desc = SystemDescription(sha256='foo',
                         types=db_type,
                         formats=db_format,
                         packages=PackageDatabase([]))
    return desc


def check_example_bag(db_type: TypeDatabase, bag: BagReader) -> None:
    typ_mavlink = db_type['mavros_msgs/Mavlink']
    assert bag.topics == {'/rosout', '/mavros/mission/waypoints',
                          '/mavlink/from', '/diagnostics', '/rosout_agg',
                          '/mavros/state'}
    assert bag.header.chunk_count == 1
    assert bag.header.conn_count == 6
    msgs = list(bag.read_messages(['/mavlink/from']))
    assert all(m.topic == '/mavlink/from' for m in msgs)
    assert all(isinstance(m.message, typ_mavlink) for m in msgs)


def test_from_file():
    db_type = load_mavros_type_db()
    fn_bag = os.path.join(DIR_TEST, 'example.bag')
    bag = BagReader(fn_bag, db_type)
    check_example_bag(db_type, bag)


def test_bag_reader_time():
    db_type = load_mavros_type_db()
    fn_bag = os.path.join(DIR_TEST, 'hello_world/non-bug.bag')
    bag = BagReader(fn_bag, db_type)
    assert bag.duration == Duration(49, 449297900)
    assert bag.time_start == Time(1556312447, 647802600)
    assert bag.time_end == Time(1556312497, 97100500)


def test_bag_writer_time():
    db_type = load_mavros_type_db()
    fn_orig = os.path.join(DIR_TEST, 'hello_world/non-bug.bag')
    reader = BagReader(fn_orig, db_type)
    messages = list(reader)

    # copy the contents of the bag
    fn_copy = tempfile.mkstemp(suffix='.bag')[1]
    try:
        writer = BagWriter(fn_copy)
        writer.write(messages)
        writer.close()

        reader_copy = BagReader(fn_copy, db_type)
        assert reader_copy.duration == reader.duration
        assert reader_copy.time_start == reader.time_start
        assert reader_copy.time_end == reader.time_end
    finally:
        os.remove(fn_copy)


def test_write():
    db_type = load_mavros_type_db()

    # load the messages from the example bag
    reader = BagReader(os.path.join(DIR_TEST, 'example.bag'), db_type)
    messages = list(reader)

    fn_bag = tempfile.mkstemp(suffix='.bag')[1]
    try:
        bag = BagWriter(fn_bag)
        bag.write(messages)
        bag.close()

        reader = BagReader(fn_bag, db_type)
        check_example_bag(db_type, reader)
    finally:
        os.remove(fn_bag)


def test_simple_write():
    db_type = load_mavros_type_db()

    # load the messages from the example bag
    reader = BagReader(os.path.join(DIR_TEST, 'simple.bag'), db_type)
    messages = list(reader)

    fn_bag = tempfile.mkstemp(suffix='.bag')[1]
    try:
        bag = BagWriter(fn_bag)
        bag.write(messages)
        bag.close()

        reader = BagReader(fn_bag, db_type)
        assert reader.header.chunk_count == 1
        assert reader.header.conn_count == 1
        assert reader.topics == {'/hello'}
        msgs = list(reader.read_messages('/hello'))
    finally:
        os.remove(fn_bag)


def test_bag_replay():
    fn_bag = os.path.join(DIR_TEST, 'hello_world/bug.bag')
    with build_hello_world() as (sut, ros):
        with ros.playback(fn_bag) as player:
            player.wait()
            popen = player._process
            out = '\n'.join(popen.stream)
            assert 'error' not in out
            assert 'Done' in out


def test_write_and_replay():
    log_to_stdout = logging.StreamHandler()
    log_to_stdout.setLevel(logging.DEBUG)
    logging.getLogger('roswire.bag.writer').addHandler(log_to_stdout)
    with build_hello_world() as (sut, ros):
        fn_bag_orig = os.path.join(DIR_TEST, 'minimal.bag')
        db_type = sut.description.types
        reader = BagReader(fn_bag_orig, db_type)
        messages = list(reader)

        fn_bag = 'temp.bag'
        writer = BagWriter(fn_bag)
        writer.write(messages)
        writer.close()

        ros.launch('/ros_ws/src/ros_tutorials/roscpp_tutorials/launch/listener.launch')
        with ros.playback(fn_bag) as player:
            player.wait()
            popen = player._process
            out = '\n'.join(popen.stream)
            print(out)
            assert 'error' not in out
            assert 'Done' in out



__all__ = ('NodeManagerProxy', 'NodeProxy')

from typing import Iterator, Set, Mapping, Optional
from urllib.parse import urlparse
import xmlrpc.client
import logging

import psutil

from .shell import ShellProxy
from ..exceptions import ROSWireException, NodeNotFoundError

logger = logging.getLogger(__name__)  # type: logging.Logger
logger.setLevel(logging.DEBUG)


class NodeProxy:
    def __init__(self,
                 name: str,
                 url_host_network: str,
                 shell: ShellProxy
                 ) -> None:
        """
        Constructs a proxy for a given name.

        Parameters:
            name: the name of the node.
            url_host_network: the URL of the node on the host network.
            shell: a shell proxy.
        """
        self.__name = name
        self.__url = url_host_network
        self.__shell = shell
        self.__pid_host: Optional[int] = None

    @property
    def api(self) -> xmlrpc.client.ServerProxy:
        """Provides access to the XML-RPC API for this node."""
        return xmlrpc.client.ServerProxy(self.url)

    @property
    def name(self) -> str:
        """The fully qualified name of this node."""
        return self.__name

    @property
    def url(self) -> str:
        """URL used to access this node from the host network."""
        return self.__url

    @property
    def pid(self) -> int:
        """The container PID of the main process for this node."""
        code, status, pid = self.api.getPid('/.roswire')
        if code != 1:
            m = f"failed to obtain PID [{self.name}]: {status} (code: {code})"
            raise ROSWireException(m)
        assert isinstance(pid, int)
        assert pid > 0
        return pid

    @property
    def pid_host(self) -> int:
        """The host PID of the main process for this node."""
        if self.__pid_host is None:
            self.__pid_host = self.__shell.local_to_host_pid(self.pid)
            assert self.__pid_host is not None
        return self.__pid_host

    def is_alive(self) -> bool:
        # TODO check start time to ensure this is the same process!
        try:
            return psutil.pid_exists(self.pid_host)
        except ROSWireException:
            return False

    def shutdown(self) -> None:
        self.__shell.execute(f'rosnode kill {self.name}')


class NodeManagerProxy(Mapping[str, NodeProxy]):
    def __init__(self,
                 host_ip_master: str,
                 api: xmlrpc.client.ServerProxy,
                 shell: ShellProxy
                 ) -> None:
        self.__host_ip_master: str = host_ip_master
        self.__api: xmlrpc.client.ServerProxy = api
        self.__shell: ShellProxy = shell

    @property
    def api(self) -> xmlrpc.client.ServerProxy:
        return self.__api

    def __get_node_names(self) -> Set[str]:
        """
        Fetches a list of the names of all active nodes.
        """
        names: Set[str] = set()
        code, status, state = self.api.getSystemState('/.roswire')
        for s in state:
            for t, l in s:
                names.update(n for n in l)
        return names

    def __len__(self) -> int:
        """
        Returns a count of the number of active nodes.
        """
        return len(self.__get_node_names())

    def __iter__(self) -> Iterator[str]:
        """
        Returns an iterator over the names of all active nodes.
        """
        yield from self.__get_node_names()

    def __getitem__(self, name: str) -> NodeProxy:
        """
        Attempts to fetch a given node.

        Raises:
            NodeNotFoundError: if there is no node with the given name.
        """
        code, status, uri_container = self.api.lookupNode('/.roswire', name)
        if code == -1:
            raise NodeNotFoundError(name)
        if code != 1:
            m = f"unexpected error when attempting to find node [{name}]: {status} (code: {code})"   # noqa: pycodestyle
            raise ROSWireException(m)

        # convert URI to host network
        port = urlparse(uri_container).port
        uri_host = f"http://{self.__host_ip_master}:{port}"
        return NodeProxy(name, uri_host, self.__shell)

    def __delitem__(self, name: str) -> None:
        """
        Shutdown and deregister a given node.

        Raises:
            NodeNotFoundError: no node found with given name.
        """
        try:
            node = self[name]
        except NodeNotFoundError:
            logger.exception("failed to delete node [%s]: not found.", name)
            raise
        node.shutdown()



import functools

import pytest

from roswire.proxy.substitution import resolve

from test_basic import build_file_and_shell_proxy


def test_resolve():
    with build_file_and_shell_proxy() as (files, shell):
        r = functools.partial(resolve, shell, files)
        assert r('$(dirname)/foo.txt', {'filename': '/foo/bar/launch/yes.xml'}) == '/foo/bar/launch/foo.txt'
        assert r('$(find tf2)/CMakeLists.txt') == '/ros_ws/src/geometry2/tf2/CMakeLists.txt'



import pytest
import attr

from roswire.definitions import (PackageDatabase, FormatDatabase, TypeDatabase,
                                 MsgFormat, Time)

from test_file import build_file_proxy
from test_bag import load_mavros_type_db


def test_build():
    with build_file_proxy() as files:
        paths = [
            '/ros_ws/src/geometry2/tf2_msgs',
            '/ros_ws/src/geometry/tf',
            '/ros_ws/src/common_msgs/geometry_msgs',
            '/ros_ws/src/std_msgs'
        ]
        db_package = PackageDatabase.from_paths(files, paths)
        db_format = FormatDatabase.build(db_package)
        db_type = TypeDatabase.build(db_format)
        assert set(db_type) == {
            'geometry_msgs/Point',
            'geometry_msgs/Point32',
            'geometry_msgs/Pose2D',
            'geometry_msgs/Quaternion',
            'geometry_msgs/Vector3',
            'std_msgs/Bool',
            'std_msgs/Byte',
            'std_msgs/Char',
            'std_msgs/ColorRGBA',
            'std_msgs/Duration',
            'std_msgs/Empty',
            'std_msgs/Float32',
            'std_msgs/Float64',
            'std_msgs/Header',
            'std_msgs/Int16',
            'std_msgs/Int32',
            'std_msgs/Int64',
            'std_msgs/Int8',
            'std_msgs/MultiArrayDimension',
            'std_msgs/String',
            'std_msgs/Time',
            'std_msgs/UInt16',
            'std_msgs/UInt32',
            'std_msgs/UInt64',
            'std_msgs/UInt8',
            'tf/FrameGraphResponse',
            'tf2_msgs/FrameGraphResponse',
            'tf2_msgs/LookupTransformGoal',
            'tf2_msgs/TF2Error',
            'geometry_msgs/Accel',
            'geometry_msgs/Inertia',
            'geometry_msgs/PointStamped',
            'geometry_msgs/Polygon',
            'geometry_msgs/Pose',
            'geometry_msgs/QuaternionStamped',
            'geometry_msgs/Transform',
            'geometry_msgs/Twist',
            'geometry_msgs/Vector3Stamped',
            'geometry_msgs/Wrench',
            'std_msgs/MultiArrayLayout',
            'geometry_msgs/AccelStamped',
            'geometry_msgs/AccelWithCovariance',
            'geometry_msgs/InertiaStamped',
            'geometry_msgs/PolygonStamped',
            'geometry_msgs/PoseArray',
            'geometry_msgs/PoseStamped',
            'geometry_msgs/PoseWithCovariance',
            'geometry_msgs/TransformStamped',
            'geometry_msgs/TwistStamped',
            'geometry_msgs/TwistWithCovariance',
            'geometry_msgs/WrenchStamped',
            'std_msgs/ByteMultiArray',
            'std_msgs/Float32MultiArray',
            'std_msgs/Float64MultiArray',
            'std_msgs/Int16MultiArray',
            'std_msgs/Int32MultiArray',
            'std_msgs/Int64MultiArray',
            'std_msgs/Int8MultiArray',
            'std_msgs/UInt16MultiArray',
            'std_msgs/UInt32MultiArray',
            'std_msgs/UInt64MultiArray',
            'std_msgs/UInt8MultiArray',
            'geometry_msgs/AccelWithCovarianceStamped',
            'geometry_msgs/PoseWithCovarianceStamped',
            'geometry_msgs/TwistWithCovarianceStamped',
            'tf/tfMessage',
            'tf2_msgs/LookupTransformResult',
            'tf2_msgs/TFMessage'
        }

        t = db_type['tf/FrameGraphResponse']
        m = t(dot_graph='foo')
        assert m.dot_graph == 'foo'

        m = t('bar')
        assert m.dot_graph == 'bar'

        with pytest.raises(TypeError):
            t(foo='bar')
        with pytest.raises(TypeError):
            t('foo', 'bar')


def test_to_and_from_dict():
    s = """
uint32 seq
time stamp
string frame_id
    """
    fmt = MsgFormat.from_string("PkgName", "MessageName", s)
    db_fmt = FormatDatabase({fmt}, {}, {})
    db_type = TypeDatabase.build(db_fmt)

    t = db_type['PkgName/MessageName']
    m = t(seq=310, stamp=Time(30, 120), frame_id='foo')
    d = {'seq': 310,
         'stamp': {'secs': 30, 'nsecs': 120},
         'frame_id': 'foo'}
    assert m.to_dict() == d
    assert db_type.from_dict(fmt, d) == m


def test_md5sum():
    db_type = load_mavros_type_db()

    def check(name: str, md5sum_expected: str) -> None:
        assert db_type[name].md5sum() == md5sum_expected

    check('std_msgs/Header', '2176decaecbce78abc3b96ef049fabed')
    check('std_msgs/Duration', '3e286caf4241d664e55f3ad380e2ae46')
    check('mavros_msgs/CommandCode', 'f7e54ea3892a961cc44c9350fdb0855e')
    check('geometry_msgs/Quaternion', 'a779879fadf0160734f906b8c19c7004')
    check('geometry_msgs/Vector3', '4a842b65f413084dc2b10fb484ea7f17')
    check('geometry_msgs/Wrench', '4f539cf138b23283b520fd271b567936')
    check('geometry_msgs/Transform', 'ac9eff44abf714214112b05d54a3cf9b')
    check('geometry_msgs/TransformStamped', 'b5764a33bfeb3588febc2682852579b0')
    check('tf/tfMessage', '94810edda583a504dfda3829e70d7eec')
    check('visualization_msgs/InteractiveMarkerPose', 'a6e6833209a196a38d798dadb02c81f8')
    check('mavros_msgs/State', '9e3d873fae342c8f48a8bd64c53d991e')
    check('mavros_msgs/Mavlink', '6dd71a38b8541fdc2de89a548c7dbc2f')
    check('rosgraph_msgs/Log', 'acffd30cd6b6de30f120938c17c593fb')



__all__ = ('Time', 'Duration', 'is_builtin', 'get_builtin')

from typing import Dict, Any, Type

import attr


@attr.s(frozen=True, slots=True)
class Time:
    secs: int = attr.ib()
    nsecs: int = attr.ib()

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> 'Time':
        return Time(d['secs'], d['nsecs'])

    def to_dict(self) -> Dict[str, int]:
        return {'secs': self.secs,
                'nsecs': self.nsecs}


@attr.s(frozen=True, slots=True)
class Duration:
    secs: int = attr.ib()
    nsecs: int = attr.ib()

    @staticmethod
    def between(start: Time, stop: Time) -> 'Duration':
        """Computes the length of time between two timestamps."""
        assert stop.secs > start.secs or stop.nsecs >= start.nsecs
        if start.secs == stop.secs:
            return Duration(0, stop.nsecs - start.nsecs)

        d_secs = stop.secs - start.secs
        if stop.nsecs >= start.nsecs:
            return Duration(d_secs, stop.nsecs - start.nsecs)
        else:
            sec_to_nsec = 1000000000  # 1 sec = 10^9 nsecs
            return Duration(d_secs - 1, stop.nsecs + sec_to_nsec - start.nsecs)

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> 'Duration':
        return Duration(d['secs'], d['nsecs'])

    def to_dict(self) -> Dict[str, int]:
        return {'secs': self.secs,
                'nsecs': self.nsecs}


_BUILTIN_TYPES: Dict[str, Type] = {
    'bool': bool,
    'int8': int,
    'uint8': int,
    'int16': int,
    'uint16': int,
    'int32': int,
    'uint32': int,
    'int64': int,
    'uint64': int,
    'float32': float,
    'float64': float,
    'string': str,
    'time': Time,
    'duration': Duration,
    'char': int,  # deprecated: alias for uint8
    'byte': int  # deprecated: alias for int8
}


def is_builtin(typ: str) -> bool:
    """Determines whether a given type is a built-in type."""
    return typ in _BUILTIN_TYPES


def get_builtin(typ: str) -> Type:
    """Returns the Python type that implements a given ROS type."""
    return _BUILTIN_TYPES[typ]



__all__ = ('ServiceProxy', 'ServiceProxyManager')

from typing import Iterator, Any, List, Set, Mapping, Optional
from urllib.parse import urlparse
import xmlrpc.client
import logging
import shlex
import json

import attr
import yaml

from .shell import ShellProxy
from .. import exceptions
from ..definitions import Message, SrvFormat, MsgFormat
from ..description import SystemDescription

logger: logging.Logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@attr.s(slots=True)
class ServiceProxy:
    name: str = attr.ib()
    url: str = attr.ib()
    format: SrvFormat = attr.ib()
    _description: SystemDescription = attr.ib()
    _shell: ShellProxy = attr.ib()

    def call(self, message: Optional[Message] = None) -> Optional[Message]:
        if not message:
            yml = '{}'
        else:
            yml = yaml.dump(message.to_dict())
        cmd = f"rosservice call {self.name} '{yml}'"
        code, output, duration = self._shell.execute(cmd)

        if code == 2:
            raise exceptions.ROSWireException('illegal service call args')
        if code != 0:
            raise exceptions.ROSWireException('unexpected error during service call')  # noqa

        fmt_response: Optional[MsgFormat] = self.format.response
        if not fmt_response:
            return None

        d = yaml.load(output)
        db_type = self._description.types
        return db_type.from_dict(fmt_response, d)


class ServiceManagerProxy(Mapping[str, ServiceProxy]):
    """
    Provides access to the registered services on a ROS graph.
    """
    def __init__(self,
                 description: SystemDescription,
                 host_ip_master: str,
                 api: xmlrpc.client.ServerProxy,
                 shell: ShellProxy
                 ) -> None:
        self.__description = description
        self.__host_ip_master = host_ip_master
        self.__api = api
        self.__shell = shell

    def __get_service_names(self) -> Set[str]:
        code, msg, state = self.__api.getSystemState('/.roswire')
        if code != 1:
            m = "an unexpected error occurred when retrieving services"
            m = f"{m}: {msg} (code: {code})"
            raise exceptions.ROSWireException(m)
        pubs, subs, services_and_providers = state
        services: Set[str] = set(s[0] for s in services_and_providers)
        return services

    def __len__(self) -> int:
        """
        The number of advertised services on this ROS graph.
        """
        return len(self.__get_service_names())

    def __iter__(self) -> Iterator[str]:
        """
        Returns an iterator over the names of all registered services.
        """
        yield from self.__get_service_names()

    def __getitem__(self, name: str) -> ServiceProxy:
        """
        Fetches a proxy for a service with a given name.

        Raises:
            ServiceNotFound: if no service is found with the given name.
        """
        code, msg, url_container = self.__api.lookupService('/.roswire', name)

        if code == -1:
            raise exceptions.ServiceNotFoundError(name)
        if code != 1:
            m = "an unexpected error occurred when retrieving services"
            m = f"{m}: {msg} (code: {code})"
            raise exceptions.ROSWireException(m)

        # convert URL to host network
        parsed = urlparse(url_container)
        url_host = f"{parsed.scheme}://{self.__host_ip_master}:{parsed.port}"

        # find the format for the service
        code, name_fmt, duration = \
            self.__shell.execute(f'rosservice type {name}')
        if code != 0:
            m = f"unable to determine type for service [{name}]"
            raise exceptions.ROSWireException(m)
        fmt = self.__description.formats.services[name_fmt]
        return ServiceProxy(name,
                            url_host,
                            fmt,
                            self.__description,
                            self.__shell)



# -*- coding: utf-8 -*-
__all__ = ('build_tuple', 'Stopwatch',)

from typing import TypeVar, Sequence, Tuple
from timeit import default_timer as timer
import warnings

T = TypeVar('T')


def build_tuple(elements: Sequence[T]) -> Tuple[T, ...]:
    """Tranforms a sequence of items to a tuple."""
    return tuple(elements)


class Stopwatch:
    def __init__(self) -> None:
        """
        Constructs a new, paused timer.
        """
        self.__offset: float = 0.0
        self.__paused: bool = True
        self.__time_start: float = 0.0

    def stop(self) -> None:
        """
        Freezes the timer.
        """
        if not self.__paused:
            self.__offset += timer() - self.__time_start
            self.__paused = True

    def start(self) -> None:
        """
        Resumes the timer.
        """
        if self.__paused:
            self.__time_start = timer()
            self.__paused = False
        else:
            warnings.warn("timer is already running")

    def reset(self) -> None:
        """
        Resets and freezes the timer.
        """
        self.__offset = 0.0
        self.__paused = True

    @property
    def paused(self) -> bool:
        """
        Returns True if this stopwatch is paused, or False if not.
        """
        return self.__paused

    @property
    def duration(self) -> float:
        """
        The number of seconds that the stopwatch has been running.
        """
        d = self.__offset
        if not self.__paused:
            d += timer() - self.__time_start
        return d
